#! /usr/bin/env python

"""
This script performs the main ingestion of data into the hstlc
filesystem and database, as well as creates output lightcurves for both
individual observations as well as 'composite' (i.e. aggregate)
observations of the same target. This script employs the following
algorithm:

    1. Configure logging
    2. Parse command line arguments
    3. Gather ``*_x1d.fits``, ``*_tag.fits``, ``*_corrtag.fits``,
       ``*_corrtag_a.fits``, and ``*_corrtag_b.fits`` files from the
       ``ingest_dir`` directory, as determined by the config file (see
       below):
        a. If both a ``*_corrtag_a.fits`` and a ``*_corrtag_b.fits``
        file exists for a given dataset, ignore the
        ``*_corrtag_b.fits`` file (as to avoid redundant extraction).
    4. For each dataset to ingest:
        a. Perform data quality checks.  If the dataset is deemed bad:
            i. Update the ``bad_data`` table in database
            ii. Remove dataset
        b. Gather metadata
        c. If dataset is a STIS dataset:
            i. Extract spectra
            ii. Rename ``*_tag.fits`` to ``*_corrtag.fits``
        d. If dataset is a COS dataset:
            i. Extract spectra (both ``*_corrtag_a.fits`` and
            ``*_corrtag_b.fits``, if necessary)
        e. Update ``metadata`` table in database
        f. Create lightcurve
        g. Update ``outputs`` table in database
        h. Create `quicklook' image
        i. Move file to appropriate location in filesystem
    5. Create composite lightcurve for each dataset in unique
       detector-targname-opt_elem-cenwave configuration

The filenames and headers of the composite lightcurves are configured
such that they can be delivered to MAST as High Level Science Products
(HLSPs), though not all composite lightcurves are delivered.

This script uses multiprocessing.  Users can set the number of cores
used via the ``num_cores`` setting in the config file (see below)


**Authors:**

    Matthew Bourque, Justin Ely

**Use:**

    This script is intended to be executed as part of the
    ``hstlc_pipeline`` shell script.  However, users can also execute
    this script via the command line as such:

    >>> ingest_hstlc [-corrtag_extract]

    ``-corrtag_extract`` (*optional*) - (Re)extract corrtag data as it
    is ingested, if provided

**Outputs:**

    (1) New and/or updated entries in the ``metadata``, ``outputs``,
        and ``bad_data`` tables in the hstlc database
    (2) ``x1d``, ``tag``, and ``corrtag`` files are moved from the
        ``ingest_dir`` directory to the appropriate directory in the
        ``filesystem_dir`` directory, as determined by the config file
        (see below)
    (3) ``*_curve.fits`` lightcurves for individual observations,
        placed in the appropriate ``output_dir`` directory as
        determined by the config file (see below)
    (4) ``hlsp_hstlc_*.fits`` lightcurves for composite observations,
        placed in the ``composite_dir`` directory , as determined by
        the config file (see below)
    (5) a log file in the ``log_dir`` directory as determined by the
        config file (see below)

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have access to the cdbs ``lref`` and ``oref``
        directories, which hold COS and STIS reference files,
        respectively
    (3) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``ingest_dir`` - The path to where files to be ingested are
          stored
        - ``filesystem_dir`` - The path to the hstlc filesystem
        - ``outputs_dir`` - The path to where hstlc output products are
          stored
        - ``composite_dir`` - The path to where hstlc composite output
          products are stored.
        - ``log_dir`` - The path to where the log file will be stored
        - ``num_cores`` - The number of cores to use during
          multiprocessing

    Other external library dependencies include:
        - ``astropy``
        - ``lightcurve``
        - ``lightcurve_pipeline``
        - ``pymysql``
        - ``matplotlib``
        - ``sqlalchemy``
"""

import argparse
import datetime
import glob
import itertools
import logging
import multiprocessing
import os
import shutil
import traceback

from astropy.io import fits
import lightcurve

from lightcurve_pipeline.utils.utils import make_directory
from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.utils import set_permissions
from lightcurve_pipeline.utils.utils import setup_logging
from lightcurve_pipeline.database.update_database import update_metadata_table
from lightcurve_pipeline.database.update_database import update_outputs_table
from lightcurve_pipeline.ingest.make_lightcurves import make_composite_lightcurves
from lightcurve_pipeline.ingest.make_lightcurves import make_individual_lightcurve
from lightcurve_pipeline.ingest.resolve_target import get_targname
from lightcurve_pipeline.quality.data_checks import dataset_ok

# Use matplotlib backend for quicklook images
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

# Set CRDS environement for reference files
os.environ['lref'] = '/grp/hst/cdbs/lref/'
os.environ['oref'] = '/grp/hst/cdbs/oref/'

# -----------------------------------------------------------------------------

def get_files_to_ingest():
    """
    Return a list of files to ingest.  Since ``corrtag_a`` and
    ``corrtab_b`` files are extracted together, the returned list must
    be void of duplicate ``corrtag`` files in order to avoid double
    extraction

    Returns
    -------
    files_to_ingest : list
        A list of full paths to files to ingest
    """

    logging.info('')
    logging.info('Gathering files to ingest')

    # Determine which files are corrtag_a or corrtag_b files
    files_to_ingest = glob.glob(os.path.join(SETTINGS['ingest_dir'], '*tag*.fits'))
    corrtag_ab_files = [item for item in files_to_ingest if 'corrtag_' in os.path.basename(item)]
    files_to_remove = []

    # Check to see if both a corrtag_a and corrtag_b file exists for a given rootname
    for corrtag_file in corrtag_ab_files:
        corrtag_dirname = os.path.dirname(corrtag_file)
        corrtag_rootname = os.path.basename(corrtag_file).split('_')[0]
        corrtag_a_file = '{}/{}_corrtag_a.fits'.format(corrtag_dirname, corrtag_rootname)
        corrtag_b_file = '{}/{}_corrtag_b.fits'.format(corrtag_dirname, corrtag_rootname)
        if os.path.exists(corrtag_a_file) and os.path.exists(corrtag_b_file):
            files_to_remove.append(corrtag_b_file)

    # If both a corrtag_a and a corrtag_b file exists, remove the corrtag_b
    for file_to_remove in set(files_to_remove):
        files_to_ingest.remove(file_to_remove)

    return files_to_ingest

# -----------------------------------------------------------------------------

def ingest(mp_args):
    """Ingests the file into the hstlc filesystem and database. Also
    produces output lightcurves

    Parameters
    ----------
    mp_args : tuple
        The multiprocessing arguments.  The zeroth value is the
        filename (i.e. the file to ingest), and the first value is the
        corrtag_extract switch (i.e. turn on/off stis corrtag
        re-extraction)
    """

    # Parse multiprocessing args
    filename = mp_args[0]
    corrtag_extract = mp_args[1]

    try:

        logging.info('Ingesting {}'.format(filename))

        # Open file
        with fits.open(filename, 'readonly') as hdulist:
            header = hdulist[0].header

            # Check that quality of the file before ingesting
            success = dataset_ok(filename)

            # Ingest the data if it is ok
            if success:
                metadata_dict, outputs_dict = make_file_dicts(filename, header)

                # If the file is a _tag STIS file, then make a corrtag
                if metadata_dict['instrume'] == 'STIS' and '_tag.fits' in filename:
                    lightcurve.stis.stis_corrtag(filename)
                    new_filename = filename.replace('_tag.fits', '_corrtag.fits')
                    metadata_dict['filename'] = os.path.basename(new_filename)

                # If the file is a corrtag STIS file, then re-extract if corrtag_extract is on
                elif metadata_dict['instrume'] == 'STIS' and '_corrtag.fits' in filename and corrtag_extract:
                    lightcurve.stis.stis_corrtag(filename)

                update_metadata_table(metadata_dict)

                success = make_individual_lightcurve(metadata_dict, outputs_dict)
                if success:
                    update_outputs_table(metadata_dict, outputs_dict)
                    #make_quicklook(outputs_dict)

                # Move file into the hstlc filesystem
                move_file(metadata_dict)

    # Track any errors that happen during processing
    except Exception as error:
        trace = 'Failed to ingest {}\n{}'.format(filename, traceback.format_exc())
        logging.critical(trace)

# -----------------------------------------------------------------------------

def make_file_dicts(filename, header):
    """Return a dictionary containing file metadata and a dictionary
    containing output product information

    Parameters
    ----------
    filename : string
        The absolute path to the file
    header : astropy.io.fits.header.Header
        The primary header of the file

    Returns
    -------
    metadata_dict : dict
        A dictionary containing metadata of the file
    outputs_dict : dict
        A dictionary containing output product information
    """

    metadata_dict = {}
    outputs_dict = {}

    # Set header keys
    metadata_dict['telescop'] = header['TELESCOP']
    metadata_dict['instrume'] = header['INSTRUME']
    metadata_dict['targname'] = get_targname(header['TARGNAME'])
    metadata_dict['cal_ver'] = header['CAL_VER']
    metadata_dict['obstype'] = header['OBSTYPE']
    metadata_dict['aperture'] = header['APERTURE']
    metadata_dict['detector'] = header['DETECTOR']
    metadata_dict['opt_elem'] = header['OPT_ELEM']

    # cenwave should be set to 0 for IMAGING OBSTYPE
    if header['OBSTYPE'] == 'SPECTROSCOPIC':
        metadata_dict['cenwave'] = header['CENWAVE']
    elif header['OBSTYPE'] == 'IMAGING':
        metadata_dict['cenwave'] = 0

    # fppos should be set to 0 for STIS data
    if header['INSTRUME'] == 'COS':
        metadata_dict['fppos'] = header['FPPOS']
    elif header['INSTRUME'] == 'STIS':
        metadata_dict['fppos'] = 0

    # Set image metadata keys
    metadata_dict['filename'] = os.path.basename(filename)
    metadata_dict['path'] = os.path.join(SETTINGS['filesystem_dir'], metadata_dict['targname'])
    metadata_dict['ingest_date'] = datetime.datetime.strftime(datetime.datetime.today(), '%Y-%m-%d')

    # Set outputs keys
    outputs_dict['individual_path'] = metadata_dict['path'].replace('filesystem', 'outputs')
    outputs_dict['individual_filename'] = '{}_curve.fits'.format(metadata_dict['filename'].split('_')[0])

    return metadata_dict, outputs_dict

# -----------------------------------------------------------------------------

def make_quicklook(outputs_dict):
    """Make a quicklook PNG of the lightcurve

    Parameters
    ----------
    outputs_dict : dict
        A dictionary containing output product information
    """

    lc_name = os.path.join(outputs_dict['individual_path'], outputs_dict['individual_filename'])
    lightcurve.io.quicklook(lc_name)
    set_permissions(lc_name.replace('.fits', '.png'))

# -----------------------------------------------------------------------------

def move_file(metadata_dict):
    """Move the file (and it's accompanying ``x1d`` file) from the
    ingest directory into the filesystem.  The parent directory to the
    ile is named afer the file's ``TARGNAME``

    Parameters
    ----------
    metadata_dict : dict
        A dictionary containing metadata of the file
    """

    # Create parent directory if necessary
    make_directory(metadata_dict['path'])

    # Move the file from ingest directory into filesystem
    src = os.path.join(SETTINGS['ingest_dir'], metadata_dict['filename'])

    # If the file is a corrtag_a/b file, then move any accompanying corrtag
    # file.  If not, move the file nominally
    src_list = []
    dst_list = []
    if 'corrtag_' in src:
        filelist = lightcurve.cos.get_both_filenames(src)
        for filename in filelist:
            src_list.append(filename)
            dst_list.append(os.path.join(metadata_dict['path'], os.path.basename(filename)))
    else:
        src_list.append(src)
        dst_list.append(os.path.join(metadata_dict['path'], metadata_dict['filename']))
    for src, dst in zip(src_list, dst_list):
        if os.path.exists(dst):
            os.remove(dst)
        if os.path.exists(src):
            shutil.move(src, dst)

    # Move the accompanying x1d file from ingest directory to filesystem
    x1d_file = '{}_x1d.fits'.format(metadata_dict['filename'].split('_')[0])
    src = os.path.join(SETTINGS['ingest_dir'], x1d_file)
    dst = os.path.join(metadata_dict['path'], x1d_file)
    if os.path.exists(src):
        if os.path.exists(dst):
            os.remove(dst)
        shutil.move(src, dst)

# -----------------------------------------------------------------------------
# ----------------------------------------------------------------------------

def parse_args():
    """Parse command line arguments. Returns ``args`` object

    Returns
    -------
    args : argparse.Namespace object
        An argparse object containing all of the added arguments
    """

    # Create help strings
    corrtag_extract_help = 'If provided, STIS corrtag re-extraction is performed.'

    # Add arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-corrtag_extract',
        dest='corrtag_extract',
        action='store_true',
        help=corrtag_extract_help)

    # Set the defaults
    parser.set_defaults(corrtag_extract=False)

    # Parse args
    args = parser.parse_args()

    return args

# -----------------------------------------------------------------------------

def main():
    """The main function of the ``ingest_hstlc`` script
    """

    # Configure logging
    module = os.path.basename(__file__).strip('.py')
    setup_logging(module)

    # Parse arguments
    args = parse_args()

    # Get list of files to ingest
    files_to_ingest = get_files_to_ingest()

    # Ingest the files using multiprocessing
    logging.info('')
    logging.info('Ingesting {} files using {} core(s)'.format(len(files_to_ingest), SETTINGS['num_cores']))
    logging.info('')
    pool = multiprocessing.Pool(processes=SETTINGS['num_cores'])
    mp_args = itertools.izip(files_to_ingest, itertools.repeat(args.corrtag_extract))
    pool.map(ingest, mp_args)
    pool.close()
    pool.join()

    # Make composite lightcurves
    make_composite_lightcurves()

    logging.info('Processing complete.')

# -----------------------------------------------------------------------------

if __name__ == '__main__':

    main()



"""This module houses several functions that are key to several modules
and scripts within the hstlc package.  Please see the individual
function documentation for more information.

**Authors:**

    Matthew Bourque

**Use:**

    This functions within this module are intended to be imported by
    the various hstlc scripts and modules, as such:

::

    from lightcurve_pipeline.utils.utils import SETTINGS
    from lightcurve_pipeline.utils.utils import insert_or_update
    from lightcurve_pipeline.utils.utils import set_permissions
    from lightcurve_pipeline.utils.utils import setup_logging
    from lightcurve_pipeline.utils.utils import make_directory

**Dependencies:**

    External library dependencies include:
        - ``astropy``
        - ``lightcurve_pipeline``
        - ``numpy``
        - ``pymyslq``
        - ``sqlalchemy``
"""

import datetime
import getpass
import grp
import logging
import os
import socket
import sys
import yaml

import astropy
import numpy
import sqlalchemy

# -----------------------------------------------------------------------------

def get_settings():
    """Return the setting information located in the configuration file
    located in the ``lightcurve_pipeline/utils/`` directory

    Returns
    -------
    data : dict
        A dictionary containing the settings present in the config.yaml
        configuration file.  Thus, the keys of this dictionary
        presumably are:

            (1) ``db_connection_string``
            (2) ``ingest_dir``
            (3) ``filesystem_dir``
            (4) ``outputs_dir``
            (5) ``composite_dir``
            (6) ``log_dir``
            (7) ``download_dir``
            (8) ``plot_dir``
            (9) ``bad_data_dir``
            (10) ``home_dir``

        The values of the keys are the user-supplied configurations
    """

    config_file = os.path.join(os.path.dirname(__file__), 'config.yaml')
    with open(config_file, 'r') as f:
        data = yaml.load(f)

    return data

SETTINGS = get_settings()

# -----------------------------------------------------------------------------

def insert_or_update(table, data, id_num):
    """
    Insert or update the given database table with the given data.
    This function performs the logic of inserting or updating an
    entry into the hstlc database; if an entry with the given
    ``id_num`` already exists, then the entry is updated, otherwise a
    new entry is inserted.

    Parameters
    ----------
    table : sqlalchemy.ext.declarative.api.DeclarativeMeta
        The table of the database to update
    data : dict
        A dictionary of the information to update.  Each key of the
        dictionary must be a column in the given table
    id_num : string
        The row ID to update.  If ``id_num`` is blank, then a new row
        is inserted instead.
    """

    from lightcurve_pipeline.database.database_interface import engine
    from lightcurve_pipeline.database.database_interface import get_session

    session = get_session()
    if id_num == '':
        engine.execute(table.__table__.insert(), data)
    else:
        session.query(table)\
            .filter(table.id == id_num)\
            .update(data)
        session.commit()
    session.close()

# -----------------------------------------------------------------------------

def set_permissions(path):
    """
    Set the permissions of the file path to hstlc permissions settings.
    The hstlc permissions settings are groupID = ``hstlc`` and
    permissions of ``rwxrwx---``.

    Parameters
    ----------
    path : string
        The path to the file
    """

    uid = os.stat(path).st_uid
    gid = grp.getgrnam("hstlc").gr_gid

    if uid == os.getuid():
        os.chown(path, uid, gid)
        os.chmod(path, 0770)

# -----------------------------------------------------------------------------

def setup_logging(module):
    """
    This function will configure the logging for the execution of the
    given module.  Logs are written out to the ``log_dir`` directory
    (as determined by the ``config.yaml`` file) with the filename
    ``<module>_<timestamp>.log``.

    Parameters
    ----------
    module : string
        The name of the module to log
    """

    # Configure logging
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M')
    filename = '{0}_{1}.log'.format(module, timestamp)
    logfile = os.path.join(SETTINGS['log_dir'], filename)
    logging.basicConfig(
        filename=logfile,
        format='%(asctime)s %(levelname)s: %(message)s',
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO)

    # Log environment information
    logging.info('User: {0}'.format(getpass.getuser()))
    logging.info('System: {0}'.format(socket.gethostname()))
    logging.info('Python Version: {0}'.format(sys.version.replace('\n', '')))
    logging.info('Python Path: {0}'.format(sys.executable))
    logging.info('Numpy Version: {0}'.format(numpy.__version__))
    logging.info('Numpy Path: {0}'.format(numpy.__path__[0]))
    logging.info('Astropy Version: {0}'.format(astropy.__version__))
    logging.info('Astropy Path: {0}'.format(astropy.__path__[0]))
    logging.info('SQLAlchemy Version: {0}'.format(sqlalchemy.__version__))
    logging.info('SQLAlchemy Path: {0}'.format(sqlalchemy.__path__[0]))

    set_permissions(logfile)

# -----------------------------------------------------------------------------

def make_directory(directory):
    """Create a directory if it doesn't already exist and set the hstlc
    permissions

    Parameters
    ----------
    directory : string
        The path to the directory
    """

    if not os.path.exists(directory):
        try:
            os.mkdir(directory)
            set_permissions(directory)
        except OSError:
            pass



#! /usr/bin/python

"""
This script retreives COS & STIS TIMETAG data from the MAST archive by
submitting XML requests.  The datasets to download is determined by
comparing the contents of the hstlc database to the contents of the
MAST database; any COS/STIS TIMETAG data that exists in MAST but does
not exist in the hstlc database is retreived.  Data is downloaded to
the ``ingest_dir`` directory determine by the config file (see below).

**Authors:**

    Matthew Bourque

**Use:**

    This script is intended to be executed via the command line as
    such:

    >>> download_hstlc

**Outputs:**

    The following filetypes are retreived (if available) and placed
    in the ``ingest_dir`` directory:

    - ``*_x1d.fits`` - 1 dimensional extracted spectra
    - ``*_tag.fits`` - STIS TIMETAG data
    - ``*_corrtag.fits`` - COS NUV TIMETAG data
    - ``*_corrtag_<a or b>.fits`` - COS FUV TIMETAG data

    Submission results are also saved to an XML file and stored in the
    ``download_dir`` directory determined by the config file (see
    below). The submission results indicate if the XML request was
    sucessful or if there were errors.

    Executing this script creates a log file in the ``log_dir``
    directory as determined by the config file (see below)

**Dependencies:**

    (1) As of early 2016, submission of XML requests to the MAST
        archive requires a special Python 2.6 environemnt with specific
        XML libraries installed.  More information can be found here:

        https://confluence.stsci.edu/display/STScISSOPublic/ArchiveXMLsubmitPKImaterial

        Additionally, ``tsql`` must be installed and the tsql executable
        must be placed in the directory ``~/freetds/bin/tsql``.  ``tsql``
        can be downloaded using freetds (http://www.freetds.org/).

    (2) Users must have access to the hstlc database

    (3) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``ingest_dir`` - The path to where the files will be stored
          after retreival
        - ``log_dir`` - The path to where the log file will be stored
        - ``download_dir`` - the path to where XML submission results
          will be stored
        - ``mast_server`` - The MAST server hostname
        - ``mast_database`` - The name of the MAST database
        - ``mast_account`` - The MAST account username
        - ``mast_password`` - The MAST account password
        - ``archive_user`` - The requester username
        - ``email`` - the requester email address
        - ``host`` - The hostname of the machine used for ftp
        - ``ftp_user`` - The username of the account of the machine
          used for ftp
        - ``dads_host`` - The hostname of the machine on which the
          MAST database resides
        - ``archive`` - The HTTPs connection hostname

    Other external library dependencies include:
        - ``lightcurve_pipeline``
"""

import glob
import datetime
import httplib
import logging
import os
import string
import getpass
import re

try:
    from urllib.request import urlopen
    from urllib.parse import urlparse, urlencode
    from http.client import HTTPSConnection
except ImportError:
    from urlparse import urlparse
    from urllib import urlopen, urlencode
    from httplib import HTTPSConnection

from lightcurve_pipeline.download.SignStsciRequest import SignStsciRequest

from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.utils import set_permissions
from lightcurve_pipeline.utils.utils import setup_logging
from lightcurve_pipeline.database.database_interface import engine
from lightcurve_pipeline.database.database_interface import session
from lightcurve_pipeline.database.database_interface import BadData
from lightcurve_pipeline.database.database_interface import Metadata

REQUEST_TEMPLATE = string.Template('\
<?xml version="1.0"?>\
 <!DOCTYPE distributionRequest SYSTEM "http://dmswww.stsci.edu/dtd/sso/distribution.dtd">\n\
   <distributionRequest Id=\'distributionRequest\'>\n\
     <head>\n\
       <requester userId="$archive_user" email="$archiveUserEmail"  source="starview"/>\n\
       <delivery>\n\
         <ftp hostName= "$ftpHost" loginName="$ftpUser" directory="$ftpDir" secure="true"/>\n\
       </delivery>\n\
       <process compression="none"/> \n\
     </head>\n\
     <body>\n\
       <include>\n\
         <select>\n\
           <suffix name=\"X1D\" />\n\
           <suffix name=\"TAG\" />\n\
           <suffix name=\"CTG\" />\n\
           <suffix name=\"CTA\" />\n\
           <suffix name=\"CTB\" />\n\
        </select> \n\
        $datasets \n\
       </include>\n\
     </body>\n\
   </distributionRequest>\n')

# -----------------------------------------------------------------------------

def build_xml_request(datasets):
    """Build the XML request for the given datasets

    Parameters
    ----------
    datasets : list
        A list of rootnames to download from MAST.

    Returns
    -------
    xml_request : string
        The XML request string.
    """

    # The list of datasets must be one string
    datasets = ''.join(['<rootname>{0}</rootname>\n'.format(rootname) for rootname in datasets])

    request_string = REQUEST_TEMPLATE.safe_substitute(
        archive_user=SETTINGS['archive_user'],
        archiveUserEmail=SETTINGS['email'],
        ftpHost=SETTINGS['host'],
        ftpDir=SETTINGS['ingest_dir'],
        ftpUser=SETTINGS['ftp_user'],
        datasets=datasets)

    xml_request = string.Template(request_string)
    xml_request = xml_request.template

    return xml_request

# -----------------------------------------------------------------------------

def get_filesystem_rootnames():
    """Return a list of the rootnames in the hstlc database.

    This list is compared to the MAST database to determine which
    datasets to download.

    Returns
    -------
    filesystem_rootnames : list
        A list of rootnames that are in the hstlc filesystem.
    """

    filesystem = session.query(Metadata.filename).all()
    filesystem_rootnames = [item[0].split('_')[0] for item in filesystem]

    return filesystem_rootnames

# -----------------------------------------------------------------------------

def get_mast_rootnames():
    """Return a list of rootnames of all COS & STIS TIMETAGE data in
    MAST.

    The following target names are ignored:
        DARK
        BIAS
        DEUTERIUM
        WAVE
        ANY
        NONE

    Returns
    -------
    mast_rootnames : list
        A list of rootnames of COS/STIS TIMETAG data in the MAST
        archive.
    """

    # Gather configuration settings
    mast_server = SETTINGS['mast_server']
    mast_database = SETTINGS['mast_database']
    mast_account = SETTINGS['mast_account']
    mast_password = SETTINGS['mast_password']

    # Build comparison date
    today = datetime.datetime.utcnow()
    today = today.strftime('%Y-%m-%d')

    # Connect to server
    print("tsql -S {0} -D '{1}' -U '{2}' -P '{3}' -t '|'".format(mast_server, mast_database, mast_account, mast_password))
    #sys.exit()
    transmit, receive = os.popen2("tsql -S {0} -D '{1}' -U '{2}' -P '{3}' -t '|'".format(mast_server, mast_database, mast_account, mast_password))

    # Build query
    query = ("SELECT assoc_member.asm_member_name,science.sci_data_set_name  "
             "FROM assoc_member "
             "JOIN science "
             "ON assoc_member.asm_asn_id = science.sci_data_set_name "
             "WHERE (science.sci_instrume = 'COS' OR science.sci_instrume = 'STIS') "
             "AND science.sci_operating_mode = 'TIME-TAG' "
             "AND science.sci_release_date < '{0}' "
             "AND science.sci_targname NOT IN ('DARK', 'BIAS', 'DEUTERIUM', 'WAVE', 'ANY', 'NONE') "
             "AND assoc_member.asm_member_type != 'PROD-FP' "
             "AND assoc_member.asm_member_type != 'PRODUCT' "
             "AND assoc_member.asm_member_type != 'EXP-GWAVE' "
             "AND assoc_member.asm_member_type != 'EXP-IWAVE' "
             "AND assoc_member.asm_member_type != 'EXP-AWAVE' "
             "AND assoc_member.asm_member_type != 'GO-WAVECAL' "
             "AND assoc_member.asm_member_type != 'AUTO-WAVECAL' "
             "\ngo\n".format(today))

    # Perform query and capture results
    transmit.write(query)
    transmit.close()
    mast_results = receive.readlines()
    receive.close()

    # Prune mast_results of unwanted information
    mast_results = mast_results[7:-2]

    mast_datasets = [item.split('|')[0].lower().strip() for item in mast_results]
    mast_associations = [item.split('|')[1].lower().strip() for item in mast_results]

    return mast_datasets, mast_associations

# -----------------------------------------------------------------------------

def save_submission_results(submission_results):
    """Save the submission results to an XML file.

    Submission results are saved in a separate XML file and stored in
    the 'download_dir' directory as determine by the config file.

    Parameters
    ----------
    submission_results : httplib object
        The submission results returned by MAST after the XML request
        is submitted.
    """

    today = datetime.datetime.strftime(datetime.datetime.now(), '%Y_%m_%d')
    xml_filename = 'result_{0}.xml'.format(today)
    xml_output_file = os.path.join(SETTINGS['download_dir'], xml_filename)
    with open(xml_output_file, 'w') as f:
        f.write(submission_results)
    set_permissions(xml_output_file)
    logging.info('\tXML file saved to: {0}'.format(xml_output_file))

# -----------------------------------------------------------------------------

def submit_xml_request(xml_request):
    """Submit the XML request to the MAST archive.

    Parameters
    ----------
    xml_request : string
        The request XML string.

    Returns
    -------
    submission_results : httplib object
        The XML request submission results.
    """

    user = os.environ.get("USER")
    home = os.environ.get("HOME")

    signer = SignStsciRequest()
    request_xml_str = signer.signRequest('{0}/.ssh/privkey.pem'.format(home), xml_request)
    params = urlencode({
        'dadshost': SETTINGS['dads_host'],
        'dadsport': 4703,
        'mission':'HST',
        'request': request_xml_str})
    headers = {"Accept": "text/html", "User-Agent":"{0}PythonScript".format(user)}
    req = httplib.HTTPSConnection(SETTINGS['archive'])
    req.request("POST", "/cgi-bin/dads.cgi", params, headers)
    response = req.getresponse().read()
    req.close()

    return response

# ------------------------------------------------------------------------------

def everything_retrieved(tracking_id):
    '''
    Check every 15 minutes to see if all submitted datasets have been
    retrieved. Based on code from J. Ely.
    Parameters:
        tracking_id : string
            A submission ID string..
    Returns:
        done : bool
            Boolean specifying is data is retrieved or not.
        killed : bool
            Boolean specifying is request was killed.
    '''

    done = False
    killed = False
#    print tracking_id
    status_url = "http://archive.stsci.edu/cgi-bin/reqstat?reqnum=={0}".format(tracking_id)
    for line in urlopen(status_url).readlines():

        if isinstance(line, bytes):
            line = line.decode()

        if "State" in line:
            if "COMPLETE" in line:
                done = True
            elif "KILLED" in line:
                killed = True
    return done, killed

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

def main():
    """The main function of the ``download_hstlc`` script
    """

    # Set up logging
    module = os.path.basename(__file__).strip('.py')
    setup_logging(module)

    # Build list of files in filesystem
    filesystem_rootnames = set(get_filesystem_rootnames())
    logging.info('{0} rootnames in filesystem.'.format(len(filesystem_rootnames)))

    # Query MAST for datasets
    mast_datasets, mast_associations = get_mast_rootnames()
    logging.info('{0} rootnames in MAST.'.format(len(mast_datasets)))
    logging.info('{0} associations in MAST.'.format(len(mast_associations)))

    # Remove bad data rootnames
    bad_rootnames = session.query(BadData.filename).all()
    bad_rootnames = set([item[0][0:9] for item in bad_rootnames])
    filesystem_rootnames = filesystem_rootnames.union(bad_rootnames)

    # Remove rootnames that already exist in ingest queue
    files_in_ingest = glob.glob(os.path.join(SETTINGS['ingest_dir'], '*_corrtag.fits'))
    files_in_ingest.extend(glob.glob(os.path.join(SETTINGS['ingest_dir'], '*_corrtag_?.fits')))
    files_in_ingest.extend(glob.glob(os.path.join(SETTINGS['ingest_dir'], '*_tag.fits')))
    rootnames_in_ingest = set([os.path.basename(item).split('_')[0] for item in files_in_ingest])
    filesystem_rootnames = filesystem_rootnames.union(rootnames_in_ingest)

    # Compare lists
    files_to_download = list({asn for asn, dset in zip(mast_associations, mast_datasets) if not dset in filesystem_rootnames})
    logging.info('{0} new associations.'.format(len(files_to_download)))

    ### import pdb; pdb.set_trace()

    # Limit number of requests to 100
    logging.info('{0} Total files found to download.'.format(len(files_to_download)))
    request_stepsize = 20

    for start in range(0, len(files_to_download), request_stepsize):
        logging.info('Downloading {}'.format(len(files_to_download[start:start+request_stepsize])))
        print('Downloading {}'.format(len(files_to_download[start:start+request_stepsize])))
        subset_to_download = files_to_download[start:start+request_stepsize]

        # Build XML request
        logging.info('Building XML request.')
        print('Building XML request.')
        xml_request = build_xml_request(subset_to_download)

        # Send request
        logging.info('Submitting XML request.')
        print('Building XML request.')
        submission_results = submit_xml_request(xml_request)

        username = getpass.getuser()
        tracking_id = re.search("("+username+"[0-9]{5})", submission_results).group()

        # Save submission results
        save_submission_results(submission_results)

        done = False
        killed = False

        while not done:
            print("waiting for files to be delivered")
            time.sleep(60)
            done, killed = everything_retrieved(tracking_id)

            if killed:
                return False

# -----------------------------------------------------------------------------

if __name__ == '__main__':

    main()



"""
This module serves as an interface for updating the various tables of
the hstlc database, either by inserting new records, or updating
existing ones

**Authors:**

    Matthew Bourque

**Use:**

    This module is intended to be imported from the various hstlc
    scripts, as such:

::

    from lightcurve_pipeline.database.update_database import update_bad_data_table
    from lightcurve_pipeline.database.update_database import update_metadata_table
    from lightcurve_pipeline.database.update_database import update_stats_table
    from lightcurve_pipeline.database.update_database import update_outputs_table

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string

    Other external library dependencies include:
        - ``pymysql``
        - ``sqlalchemy``
        - ``lightcurve_pipeline``
"""

import datetime
import logging
import os

from lightcurve_pipeline.database.database_interface import get_session
from lightcurve_pipeline.database.database_interface import Metadata
from lightcurve_pipeline.database.database_interface import Outputs
from lightcurve_pipeline.database.database_interface import BadData
from lightcurve_pipeline.database.database_interface import Stats
from lightcurve_pipeline.utils.utils import insert_or_update

# -----------------------------------------------------------------------------

def update_bad_data_table(filename, reason):
    """Insert or update a record pertaining to the filename in the
    ``bad_data`` table

    Parameters
    ----------
    filename : string
        The filename of the file
    reason : string
        The reason that the data is bad. Can either be ``No events``,
        ``Bad EXPFLAG``, ``Non-linear time``, ``Singular event``,
        ``Bad Proposal``, or ``Short Exposure``.
    """

    # Build dictionary containing data to store
    bad_data_dict = {}
    bad_data_dict['filename'] = filename
    bad_data_dict['ingest_date'] = datetime.datetime.strftime(
        datetime.datetime.today(), '%Y-%m-%d')
    bad_data_dict['reason'] = reason

    # Get the id of the record, if it exists
    session = get_session()
    query = session.query(BadData.id)\
        .filter(BadData.filename == filename).all()
    if query == []:
        id_num = ''
    else:
        id_num = query[0][0]
    session.close()

    # If id doesn't exist then instert.  If id exists, then update
    insert_or_update(BadData, bad_data_dict, id_num)

# -----------------------------------------------------------------------------

def update_metadata_table(metadata_dict):
    """Insert or update a record in the metadata table containing the
    ``metadata_dict`` information

    Parameters
    ----------
    metadata_dict : dict
        A dictionary containing metadata of the file.  Each key of the
        ``metadata_dict`` corresponds to a column in the matadata table
        of the database.
    """

    # Get the id of the record, if it exists
    session = get_session()
    query = session.query(Metadata.id)\
        .filter(Metadata.filename == metadata_dict['filename']).all()
    if query == []:
        id_num = ''
    else:
        id_num = query[0][0]
    session.close()

    # If id doesn't exist then insert. If id exsits, then update
    insert_or_update(Metadata, metadata_dict, id_num)

# -----------------------------------------------------------------------------

def update_stats_table(stats_dict, dataset):
    """Insert or update a record in the stats table for the given
    dataset containing the lightcurve product statistics given in the
    ``stats_dict``

    Parameters
    ----------
    stats_dict : dict
        A dictionary containing the lightcurve statistics. Each key of
        ``stats_dict`` corresponds to a column in the stats table of
        the database.
    dataset : string
        The path to the lightcurve product
    """

    # Get the id of the record, if it exists
    session = get_session()
    query = session.query(Stats.id)\
        .filter(Stats.lightcurve_filename == os.path.basename(dataset)).all()
    if query == []:
        id_num = ''
    else:
        id_num = query[0][0]
    session.close()

    # If id doesn't exist then instert.  If id exists, then update
    insert_or_update(Stats, stats_dict, id_num)

# -----------------------------------------------------------------------------

def update_outputs_table(metadata_dict, outputs_dict):
    """Insert or update a record in the outputs table containing
    output product information

    Parameters
    ----------
    metadata_dict : dict
        A dictionary containing metadata of the file.  Each key of the
        ``metadata_dict`` corresponds to a column in the matadata table
        of the database.
    outputs_dict : dict
        A dictionary containing output product information.  Each key
        of the ``outputs_dict`` corresponds to a column in the outputs
        table of the database.
    """

    # Get the metadata_id
    session = get_session()
    metadata_id_query = session.query(Metadata.id)\
        .filter(Metadata.filename == metadata_dict['filename']).all()
    metadata_id = metadata_id_query[0][0]
    outputs_dict['metadata_id'] = metadata_id

    # Get the id of the outputs record, if it exists
    id_query = session.query(Outputs.id)\
        .join(Metadata)\
        .filter(Metadata.filename == metadata_dict['filename']).all()
    if id_query == []:
        id_num = ''
    else:
        id_num = id_query[0][0]
    session.close()

    # If id doesn't exist then insert. If id exsits, then update
    insert_or_update(Outputs, outputs_dict, id_num)



# Copyright (C) 2014 Association of Universities for Research in Astronomy (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    3. The name of AURA and its representatives may not be used to
#      endorse or promote products derived from this software without
#      specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#
# $Log: SignStsciRequest.py,v $
# Revision 1.6  2015/06/29 19:32:17  fred
# PR 80802 remove requirement that xml work in python
#
# Revision 1.5  2014/08/14 19:11:59  fred
# PR 75462 change class name, add rcs log header
#

usexml=True
try:
  import xmlsec
  import libxml2
except ImportError:
  usexml=False
import urllib2
import urllib  # urllib2 does not define urlencode, you have to get it here.
import warnings
import re

class SignStsciRequest:

    def __init__(self):
        self.init()

    @staticmethod
    def init():
        global usexml
        if not usexml:
            return
        try:
          # Init libxml library
          libxml2.initParser()
          libxml2.substituteEntitiesDefault(1)
          # Init xmlsec library
          assert xmlsec.init() >= 0, "Error: xmlsec initialization failed."
          # Check loaded library version
          assert xmlsec.checkVersion() == 1, "Error: loaded xmlsec library version is not compatible."
          # Init crypto library
          assert xmlsec.cryptoAppInit(None) >= 0, "Error: crypto initialization failed."
          # Init xmlsec-crypto library
          assert xmlsec.cryptoInit() >= 0, "Error: xmlsec-crypto initialization failed."
        except:
          usexml=False

    @staticmethod
    def signRequest(file, request, dtd="http://dmswww.stsci.edu/dtd/sso/distribution.dtd", cgi="https://archive.stsci.edu/cgi-bin/dads.cgi", mission='HST'):
     global usexml
     if usexml:
      try:
        keysmngr = xmlsec.KeysMngr()
        if keysmngr is None:
            raise RuntimeError, "Error: failed to create keys manager."
        if xmlsec.cryptoAppDefaultKeysMngrInit(keysmngr) < 0:
            keysmngr.destroy()
            raise RuntimeError, "Error: failed to initialize keys manager."

        key = xmlsec.cryptoAppKeyLoad(filename = file, pwd = None,
           format = xmlsec.KeyDataFormatPem, pwdCallback = None,
           pwdCallbackCtx = None)
        if xmlsec.cryptoAppDefaultKeysMngrAdoptKey(keysmngr, key) < 0:
            keysmngr.destroy()
            raise RuntimeError, "Error: failed to load key into keys manager" 

        dsig_ctx = xmlsec.DSigCtx(keysmngr)

        # Match the dtd and replace it.

        pat = re.compile("(^.*<!DOCTYPE.*distributionRequest[^>]*SYSTEM[ \t]*\")([^\"]*)(\"[^>]*>.*$)", re.DOTALL)
        m = pat.match(request)

        request = m.group(1)+dtd+m.group(3)

        ctxt = libxml2.createMemoryParserCtxt(request, len(request))
        ctxt.validate(1)
        ctxt.parseDocument()
        doc = ctxt.doc()

        if doc is None or doc.getRootElement() is None:
            keysmngr.destroy()
            raise RuntimeError, "Error: unable to parse XML data"

        # find the XML-DSig start node
        node = xmlsec.findNode(doc.getRootElement(), xmlsec.NodeSignature,
                       xmlsec.DSigNs)
        if node is None:
            fragment = libxml2.parseDoc("""<Signature xmlns="http://www.w3.org/2000/09/xmldsig#">
<SignedInfo>
  <CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"/>
  <SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"/>
  <Reference URI="#distributionRequest">
    <Transforms>
      <Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>
      <Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#WithComments" />
    </Transforms>
    <DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>
    <DigestValue></DigestValue>
  </Reference>
</SignedInfo>
<SignatureValue></SignatureValue>
<KeyInfo>
 <KeyValue><RSAKeyValue>
  <Modulus></Modulus>
  <Exponent></Exponent>
 </RSAKeyValue></KeyValue> 
</KeyInfo>
</Signature>
""")
            # remove the xml header on the front of the document fragment
            fragment = fragment.getRootElement()
            # getElementsByTagName doesn't exist here for some reason, have to use xpath
            ctxt = doc.xpathNewContext()
            nodeList = ctxt.xpathEval("/distributionRequest")
            for child in nodeList:
                child.addChild(fragment)
                if child.prop('Id') == None:
                    child.setProp('Id', 'distributionRequest')
            node = xmlsec.findNode(doc.getRootElement(), xmlsec.NodeSignature,
                       xmlsec.DSigNs)
	    

        # Remove passwords

        ctxt = doc.xpathNewContext()
        nodeList = ctxt.xpathEval('//requester')
        for child in nodeList:
            child.unsetProp('archivePassword')

        nodeList = ctxt.xpathEval('//ftp')
        for child in nodeList:
            if child.hasProp('loginPassword'):
                child.unsetProp('loginPassword')
                warnings.warn('ftp password is not allowed in user requests.')

        # Sign the template, or resign existing block
        status = dsig_ctx.sign(node)
        output = str(doc)
        doc.freeDoc()
        keysmngr.destroy()
        if status < 0:
            raise RuntimeError, "Error: signature failed"
        return output
      except:
        usexml=False
        return signRequest(file, request, dtd, cgi)
     else:
        values = {'request' : request, 'privatekey' : open(file).read(), 'mission' : mission }
	data = urllib.urlencode(values)
        req = urllib2.Request(url=cgi, data=data)
	f = urllib2.urlopen(req)
        return f.read()

    def __del__(self):
        self.cleanup()

    @staticmethod
    def cleanup():
      if usexml:
        # Shutdown xmlsec-crypto library
        xmlsec.cryptoShutdown()
        # Shutdown crypto library
        xmlsec.cryptoAppShutdown()
        # Shutdown xmlsec library
        xmlsec.shutdown()
        # Shutdown LibXML2
        libxml2.cleanupParser()



"""
Perform data quality checks for the given dataset.  The dataset is
checked for a number of issues, which include:

(1) A non-normal ``EXPFLAG`` - indicating that something went wrong
    during the observation
(2) A non-linear time column in which time does not progress linearly
    through the ``TIME`` column in the dataset
(3) A dataset not having any events
(4) A dataset in which all events occur at a single time
(5) A dataset that is part of a problematic proposal
(6) A dataset with an exposure time that is too short

Datasets that do not pass these checks are moved to the
``bad_data_dir``, as determined by the config file (see below)

**Authors:**

    Justin Ely, Matthew Bourque

**Use:**

    This module is intended to be imported and used by the
    ``ingest_hstlc`` script as such:

::

    from lightcurve_pipeline.quality.data_checks import dataset_ok
    dataset_ok(dataset)

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``bad_data_dir`` - The directory in which bad data files are
          stored

    Other external library dependencies include:
        - ``astropy``
        - ``lightcurve_pipeline``
        - ``pymysql``
        - ``sqlalchemy``
"""

import inspect
import logging
import os
import shutil

from astropy.io import fits

from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.utils import set_permissions
from lightcurve_pipeline.database.update_database import update_bad_data_table

#-------------------------------------------------------------------------------

def dataset_ok(filename, move=True):
    """Perform quality check on the given dataset, and update the
    ``bad_data`` table and move the dataset to the ``bad_data``
    directory if it doesn't pass

    Parameters
    ----------
    filename : string
        The full path to the dataset
    move : bool, optional
        Whether or not to update the ``bad_data`` table and move the
        file

    Returns
    -------
    ``True`` if the dataset passes all of the quality checks, ``False``
    if it doesn't.
    """

    all_functions = [value for key, value in inspect.currentframe().f_globals.iteritems() if key.startswith('check_')]

    with fits.open(filename) as hdu:

        for func in all_functions:
            success, reason = func(hdu)
            if not success:
                if move:
                    logging.info('\tBad data for {}: {}'.format(filename, reason))
                    update_bad_data_table(os.path.basename(filename), reason)
                    move_file(filename)
                return False

    return True

#-------------------------------------------------------------------------------

def move_file(filename):
    """Move the given dataset to the ``bad_data`` directory

    Parameters
    ----------
    filename : string
        The full path to the dataset
    """

    dst = os.path.join(SETTINGS['bad_data_dir'], os.path.basename(filename))
    if os.path.exists(dst):
        os.remove(dst)
    shutil.move(filename, dst)
    set_permissions(dst)

#-------------------------------------------------------------------------------

def check_expflag(hdu):
    """Check that the ``EXPFLAG`` keyword is ``NORMAL``

    Parameters
    ----------
    hdu : astropy.io.fits.hdu.hdulist.HDUList
        The hdulist of the dataset

    Returns
    -------
    success : boolean
        ``True`` if the ``EXPFLAG`` is ``NORMAL``, ``False`` otherwise
    reason : string
        An empty string if success is ``True``, ``Bad EXPFLAG``
        otherwise
    """

    if hdu[1].header['EXPFLAG'] != 'NORMAL':
        return False, 'Bad EXPFLAG'

    return True, ''

#-------------------------------------------------------------------------------

def check_linear(hdu):
    """Check that the time column linearly progresses

    Parameters
    ----------
    hdu : astropy.io.fits.hdu.hdulist.HDUList
        The hdulist of the dataset

    Returns
    -------
    success : boolean
        ``True`` if time progressing linearly, ``False`` otherwise
    reason : string
        An empty string if success is ``True``, ``Non-linear time``
        otherwise
    """

    time_data = hdu[1].data['time']
    last = time_data[0]
    for val in time_data[1:]:
        if not val >= last:
            return False, 'Non-linear time'
        last = val

    return True, ''

#-------------------------------------------------------------------------------

def check_no_events(hdu):
    """Check that the dataset has events

    Parameters
    ----------
    hdu : astropy.io.fits.hdu.hdulist.HDUList
        The hdulist of the dataset

    Returns
    -------
    success : boolean
        ``True`` if the dataset has events, ``False`` otherwise
    reason : string
        An empty string if success is ``True``, ``No events``
        otherwise
    """

    if len(hdu[1].data) == 0:
        return False, 'No events'

    return True, ''

#-------------------------------------------------------------------------------

def check_not_singular(hdu):
    """Check that the events in the dataset are not from a single time

    Parameters
    ----------
    hdu : astropy.io.fits.hdu.hdulist.HDUList
        The hdulist of the dataset

    Returns
    -------
    success : boolean
        ``True`` if events are not from a single time, ``False``
        otherwise
    reason : string
        An empty string if success is ``True``, ``Singular event``
        otherwise
    """

    time_data = hdu[1].data['time']
    if len(set(time_data)) == 1:
        return False, 'Singular event'

    return True, ''

#-------------------------------------------------------------------------------

def check_bad_proposal(hdu):
    """Check that the proposal ID is not in a list of known 'bad'
    programs.  Programs can be bad for a number of reasons, typically
    because of specialized calibration purposes like focus sweeps or
    high-voltage tests.

    Parameters
    ----------
    hdu : astropy.io.fits.hdu.hdulist.HDUList
        The hdulist of the dataset

    Returns
    -------
    success : boolean
        ``True`` if events are not from a known bad proposal, ``False``
        otherwise
    reason : string
        An empty string if success is ``True``, ``Bad Proposal``
        otherwise
    """

    #-- 13635, 'FUV Focus Sweep Enabling Program for COS at LP3 (LENA2)  ???
    known_issue_proposals = {}

    if hdu[0].header['PROPOSID'] in known_issue_proposals:
        return False, 'Bad Proposal'

    return True, ''

#-------------------------------------------------------------------------------

def check_exptime(hdu):
    """Check that the dataset exptime is not too short.  The threshold
    is initially set to 1 second to filter out a small subset of very
    short exposures.

    Parameters
    ----------
    hdu : astropy.io.fits.hdu.hdulist.HDUList
        The hdulist of the dataset

    Returns
    -------
    success : boolean
        ``True`` if exptime greater than threshold, ``False`` otherwise
    reason : string
        An empty string if success is ``True``, ``Short Exposure``
        otherwise
    """

    if hdu[1].header['EXPTIME'] < 1:
        return False, 'Short Exposure'

    return True, ''

#-------------------------------------------------------------------------------



#! /usr/bin/env python

"""Reset all or specific tables in the hstlc database

**Authors:**

    Matthew Bourque

**Use:**

    This script is intended to be executed via the command line as
    such:

    >>> reset_hstlc_database [table]

    ``table`` (*optional*) - Reset the specific table given. Can be any
    valid table that exists in the hstlc database, ``all`` in which all
    tables will be reset, or ``production`` in which only the
    ``metadata``, ``outputs``, and ``stats`` tables will be reset.  If
    an argument is not provided, the default value of ``production`` is
    used.

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``home_dir`` - The home hstlc directory, where the
          ``bad_data`` table will be stored in a text file

    Other external library dependencies include:
        - ``lightcurve_pipeline``
        - ``pymysql``
        - ``sqlalchemy``
"""

from __future__ import print_function

import argparse
import inspect
import os
import pickle
import sys

from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.database import database_interface
from lightcurve_pipeline.database.database_interface import session
from lightcurve_pipeline.database.database_interface import engine
from lightcurve_pipeline.database.database_interface import BadData

# -----------------------------------------------------------------------------

def get_valid_tables():
    """Return a list of table names in the hstlc database

    Returns
    -------
    tables : list
        A list of hstlc table names
    """

    tables = []
    classmembers = inspect.getmembers(database_interface, inspect.isclass)
    for classmember in classmembers:
        try:
            tables.append(classmember[1].__tablename__)
        except AttributeError:
            pass

    return tables

# -----------------------------------------------------------------------------

def parse_args():
    """Parse command line arguments

    Returns
    -------
    args : argparse object
        An argparse object containing the arguments
    """

    reset_table_help = ('The table to reset. Can be any valid database table,'
        '"all" to reset all tables, or "production" to reset only the '
        'production tables.  The default option is "production".  The '
        'production tables constsit of: Metadata, Outputs, and Stats.')

    parser = argparse.ArgumentParser()
    parser.add_argument('reset_table', action='store', nargs='?', type=str,
        default='production', help=reset_table_help)
    args = parser.parse_args()

    # Make sure the argument is a valid option
    valid_options = get_valid_tables()
    valid_options.append('all')
    valid_options.append('production')
    explanation = '{} is not a valid table.'.format(args.reset_table)
    args.reset_table = args.reset_table.lower()
    assert args.reset_table in valid_options, explanation

    return args

# -----------------------------------------------------------------------------

def rebuild_production_tables():
    """Rebuild the ``prodction`` tables of the hstlc database, which
    consist of the ``metadata``, ``outputs``, and ``stats`` tables.
    The ``bad_data`` table is treated separately;  Since the
    ``bad_data`` table cannot easily be reconstructed (since bad data
    is not necessarily re-ingested), the data within the table is
    written out to a text file and re-ingested after the database is
    reset.  This essentially results in a reset database for the
    production tables, but the bad data table effectively remains
    untouched.
    """

    # # Write the contents of the bad_data table out to a text file
    pickle_file = os.path.join(SETTINGS['home_dir'], 'bad_data.pck')
    query = session.query(BadData).all()
    session.close()
    dict_list = [result.__dict__ for result in query]
    with open(pickle_file, 'w') as f:
        pickle.dump(dict_list, f)

    # Reset the database
    database_interface.base.metadata.drop_all()
    database_interface.base.metadata.create_all()

    # Rebuild the bad_data table
    with open(pickle_file, 'rb') as f:
        dict_list = pickle.load(f)
    for row in dict_list:
        engine.execute(BadData.__table__.insert(), row)

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

def main():
    """The main function of the ``reset_hstlc_database`` script
    """

    # Parse arguments
    args = parse_args()

    # Give user a second chance
    prompt = ('About to reset {} table(s) for database instance {}. '
        'Do you wish to proceed? (y/n)\n'.format(args.reset_table,
        SETTINGS['db_connection_string']))

    response = raw_input(prompt)

    if response.lower() == 'y':
        print('Resetting {} table(s)'.format(args.reset_table))

        if args.reset_table == 'all':
            database_interface.base.metadata.drop_all()
            database_interface.base.metadata.create_all()

        elif args.reset_table == 'production':
            rebuild_production_tables()

        else:
            database_interface.base.metadata.tables[args.reset_table].drop()
            database_interface.base.metadata.tables[args.reset_table].create()

# -----------------------------------------------------------------------------

if __name__ == '__main__':

    main()



"""
Generate lomb-scargle periodogram statistics.  The periodogram
statistics are used in the ``stats`` table in the hstlc database as
well as the periodogram plots generated by the ``make_hstlc_plots``
script.

**Authors:**

    Matthew Bourque

**Use:**

    This module is intended to be imported and used by the
    ``build_stats_table`` script as such:

::

    from lightcurve_pipeline.utils.periodogram_stats import get_periodogram_stats
    periods, power, mean, three_sigma, significant_periods, significant_powers = get_periodogram_stats(dataset, freq_space)

**Dependencies:**

    External library dependencies include:
        - ``astropy``
        - ``numpy``
        - ``scipy``
"""

from astropy.io import fits
import numpy as np
from scipy import signal

# -----------------------------------------------------------------------------

def get_periodogram_stats(dataset, freq_space):
    """Find significant periods from the given dataset and frequency
    space using a lomb-scargle periodogram.

    Parameters
    ----------
    dataset : string
        The path to the lightcurve product.
    freq_space : string
        Can either be ``short``, ``med``, or ``long``.  This defines
        the frequency space to look for significant periods.  ``short``
        is defined as the range (``STEPSIZE``, 10 minutes), ``med`` is
        (10 minutes, 1 hour), and ``long`` is (1 hour, 10 hours).

    Returns
    -------
    periods : numpy array
        An array of the periods to check
    power : numpy array
        An array of the lomb-scargle powers corresponding to each
        period
    mean : float
        The mean of the lomb-scargle powers
    std : float
        The standard deviation of the lomb-scargle powers
    three_sigma : float
        Three standard deviations above the mean of the lomb-scargle
        powers
    significant_periods : list
        A list of the periods that have powers greater than 3-sigma
        about the mean
    significant_powers : list
        A list of the lomb-scargle powers that are greater than 3-sigma
        about the mean
    """

    # Get the data
    hdulist = fits.open(dataset, mode='readonly')
    counts = hdulist[1].data['net']
    times = hdulist[1].data['mjd']

    # Make sure the byte-order is native
    counts = counts.byteswap().newbyteorder()
    times = times.byteswap().newbyteorder()

    # Define frequency space (in days)
    short_freq = (hdulist[0].header['STEPSIZE'] / (60. * 60. * 24.)) # Step size
    med_freq = (10. / (60. * 24.)) # 10 minutes
    long_freq = 1. / 24. # 1 hour
    max_freq = 10. / 24. # 10 hours

    if freq_space == 'short':
        periods = np.linspace(short_freq, med_freq, len(times))
    elif freq_space == 'med':
        periods = np.linspace(med_freq, long_freq, len(times))
    elif freq_space == 'long':
        periods = np.linspace(long_freq, max_freq, len(times))

    ang_freqs = 2 * np.pi / periods
    power = signal.lombscargle(np.asarray(times), np.asarray(counts) - np.asarray(counts).mean(), ang_freqs)
    power *= 2 / (len(times) * np.asarray(counts).std() ** 2)
    maxima = signal.argrelextrema(data=power, comparator=np.greater, order=1)[0]
    mean = np.mean(power)
    std = np.std(power)
    three_sigma = mean + (3 * std)
    above_three_sigma = np.where(power > three_sigma)[0]
    significant_indices = [index for index in maxima if index in above_three_sigma]
    significant_periods = periods[significant_indices]
    significant_powers = power[significant_indices]

    return periods, power, mean, three_sigma, significant_periods, significant_powers



"""
This module contains functions that attempt to resolve target names
(i.e. ``TARGNAME``) to a more common option, if possible.  The method
for doing this is as follows:

    (1) Look up the ``targname`` from the hard-coded ``targname_dict``
        dictionary. If it exists, then use that ``targname``
    (2) If no dictionary entry exists, look up the ``targname`` in the
        CDS web service[1]
    (3) If the CDS web service returns resolved target names, and one
        of those target names already exists in the ``metadata`` table,
        then use that ``targname``
    (4) If the ``targname`` cannot be resolved through any of these
        steps, then use the original ``targname``

The hard-coded ``targname_dict`` dictionary resides in the
``utils.targname_dict`` module.

**Authors:**

    Justin Ely, Matthew Bourque

**Use:**

    This module is intended to be imported from and used by the
    ``ingest_hstlc`` script as such:

::

    from lightcurve_pipeline.ingest.resolve_target import get_targname
    get_targname(targname)

**Dependencies:**

    (1) Users must have access to the CDS web service
    (2) Users must have access to the hstlc database
    (3) Users must have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string

    Other external library dependencies include:
        - ``pymysql``
        - ``sqlalchemy``
        - ``lightcurve``
        - ``lightcurve_pipeline``

**References:**

    [1] Centre de Donnees astronomiques de Strasbourg
    (http://cdsweb.u-strasbg.fr/)
"""

import urllib2
from xml.dom import minidom

from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.targname_dict import targname_dict
from lightcurve_pipeline.database.database_interface import get_session
from lightcurve_pipeline.database.database_interface import Metadata

#------------------------------------------------------------------------------

def get_targname(targname):
    """Resolve the ``targname`` to a better option, if available.  If
    the ``targname`` cannot be resolved, the original ``targname`` is
    returned.

    Parameters
    ----------
    targname : str
        The name of the target

    Returns
    -------
    new_targname: str
        The resolved target name
    """

    # Assume at first that the targname can't be resolved
    new_targname = targname

    # Try to resolve the target via the target_dict
    if targname in targname_dict.keys():
        if len(targname_dict[targname]):
            new_targname = targname_dict[targname]

    # Try to resolve the target name with the online service
    try:
        targname_set = resolve(targname)
    except:
        targname_set = set()

    if len(targname_set):

        # For each resolved target name, check to see if it's already
        # in the database.  If it is, then use that one.
        session = get_session()
        targnames_in_db = session.query(Metadata.targname).all()
        targnames_in_db = [item[0] for item in targnames_in_db]
        for item in targname_set:
            if item in targnames_in_db:
                new_targname = item
        session.close()

    return new_targname

#------------------------------------------------------------------------------

def resolve(targname):
    """Resolve target name via the CDS web service

    Parameters
    ----------
    targname : str
        The name of the target

    Returns
    -------
    other_names : set
        set of resolved other names
    """

    web_string = 'http://cdsweb.u-strasbg.fr/cgi-bin/nph-sesame/-oxpI?{}'.format(targname)

    xmldoc = minidom.parse(urllib2.urlopen(web_string))
    itemlist = xmldoc.getElementsByTagName('alias')

    other_names = [str(item.childNodes[0].data) for item in itemlist]

    return set(other_names)



#! /usr/bin/env python

"""
Reset the hstlc filesystem by moving files back into the ingestion
directory.  Files are moved from the ``filesystem_dir`` directory to
the ``ingest_dir`` directory, as determined by the config file (see
below).  Additionally, output products located in the ``outputs_dir``
directory, as determined by the config file (see below) are removed.

**Authors:**

    Matthew Bourque

**Use:**

    This script is intended to be executed via the command line as
    such:

    >>> reset_hstlc_filesystem

**Dependencies:**

    Users must have a ``config.yaml`` file located in the
    ``lightcurve_pipeline/utils/`` directory with the following
    keys:

        - ``ingest_dir`` - The path to where files to be ingested are
          stored
        - ``filesystem_dir`` - The path to the hstlc filesystem
        - ``outputs_dir`` - The path to where hstlc output products are
          stored

    Other external library dependencies include:
        - ``lightcurve_pipeline``
"""

from __future__ import print_function

import glob
import os
import shutil

from lightcurve_pipeline.utils.utils import SETTINGS

# -----------------------------------------------------------------------------

def move_files_to_ingest():
    """
    Move files from filesystem back to the ingestion directory. If the
    file already exists in the ingest directory, the file is removed
    rather than moved.
    """

    # Gather files in the filesystem directory
    filelist = glob.glob(os.path.join(SETTINGS['filesystem_dir'], '*/*.fits'))
    for filename in filelist:

        dst = os.path.join(SETTINGS['ingest_dir'], os.path.basename(filename))

        # Move the file if it doesn't already exist, otherwise simply remove it
        if not os.path.exists(dst):
            shutil.move(filename, SETTINGS['ingest_dir'])
            print('\tMoved {} to {}'.format(filename, SETTINGS['ingest_dir']))
        else:
            os.remove(filename)

# -----------------------------------------------------------------------------

def remove_output_directories():
    """
    Remove all output products and output directories
    """

    directories = glob.glob(os.path.join(SETTINGS['outputs_dir'], '*'))
    for directory in directories:
        try:
            shutil.rmtree(directory)
            print('\tRemoved directory {}'.format(directory))
        except:
            print('Could not remove {}'.format(directory))

# -----------------------------------------------------------------------------

def remove_filesystem_directories():
    """
    Remove parent directories from the filesystem if they are empty
    """

    dirlist = glob.glob(os.path.join(SETTINGS['filesystem_dir'], '*'))
    for directory in dirlist:
        try:
            os.rmdir(directory)
            print('\tRemoved directory {}'.format(directory))
        except OSError:
            print('Could not remove {}'.format(directory))

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

def main():
    """The main function of the ``reset_hstlc_filesystem`` script
    """

    # Give the user a second chance
    prompt = 'About to reset filesystem {}. '.format(SETTINGS['filesystem_dir'])
    prompt += 'Do you wish to proceed? (y/n)'
    response = raw_input(prompt)

    if response.lower() == 'y':
        print('Resetting filesystem')
        move_files_to_ingest()
        remove_filesystem_directories()
        remove_output_directories()

# -----------------------------------------------------------------------------

if __name__ == '__main__':

    main()



from setuptools import setup
from setuptools import find_packages

# Command line scripts
scripts = ['reset_hstlc_filesystem = lightcurve_pipeline.scripts.reset_hstlc_filesystem:main',
           'reset_hstlc_database = lightcurve_pipeline.scripts.reset_hstlc_database:main',
           'download_hstlc = lightcurve_pipeline.scripts.download_hstlc:main',
           'ingest_hstlc = lightcurve_pipeline.scripts.ingest_hstlc:main',
           'build_stats_table = lightcurve_pipeline.scripts.build_stats_table:main',
           'make_hstlc_plots = lightcurve_pipeline.scripts.make_hstlc_plots:main']
entry_points = {}
entry_points['console_scripts'] = scripts

setup(
    name = 'lightcurve-pipeline',
    description = 'Create lightcurves from HST/COS and HST/STIS data',
    url = 'https://github.com/justincely/lightcurve_pipeline.git',
    author = 'Matthew Bourque',
    author_email = 'bourque@stsci.edu',
    keywords = ['astronomy'],
    classifiers = ['Programming Language :: Python',
                   'Development Status :: 1 - Planning',
                   'Intended Audience :: Science/Research',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Topic :: Software Development :: Libraries :: Python Modules'],
    packages = find_packages(),
    requires = ['numpy', 'scipy', 'astropy', 'sqlalchemy', 'pyyaml', 'matplotlib', 'bokeh'],
    data_files = [('lightcurve_pipeline/utils/', ['lightcurve_pipeline/utils/config.yaml'])],
    scripts = ['scripts/hstlc_pipeline'],
    entry_points = entry_points,
    version = 1.0
    )



#! /usr/bin/env python

"""
Create various plots that deal with the hstlc filesystem, database,
and output products. This script uses multiprocessing.  Users can set
the number of cores used via the ``num_cores`` setting in the config
file (see below).

**Authors:**

    Justin Ely, Matthew Bourque

**Use:**

    This script is intended to be executed as part of the
    ``hstlc_pipeline`` shell script.  However, users can also execute
    this script via the command line as such:

    >>> make_hstlc_plots

**Outputs:**

    (1) ``hlsp_hstlc_*.png`` static lightcurve plots for each composite
        lightcurve, placed in the ``composite_dir`` directory, as
        determined by the config file (see below)
    (2) ``hlsp_hstlc_*.html`` bokeh plots showing a 'dashboard' of
        various plots for each composite lightcurve, placed in the
        ``composite_dir`` directory, as determined by the config file
        (see below)
    (3) ``interesting_hstlc.html``, ``boring_hstlc.html``, and
        ``null_hstlc.html`` 'exploratory' tables, which are sortable
        tables that display statistics and plots for each dataset,
        placed in the ``plot_dir`` directory, as determined by the
        config file (see below)
    (4) ``exptime_histogram.html`` - A histrogram showing the
        cumulative exposure time by target in the form of a bokeh plot,
        placed in the ``plot_dir`` directory, as determined by the
        config file (see below)
    (5) ``pie_config_cos_fuv.html``, ``pie_config_cos_nuv.html``,
        ``pie_config_stis_fuv.html``, and ``pie_config_stis_nuv.html``
        'configuration' pie charts that show the breakdown of datasets
        by grating/cenwave for each instrument/detector combination,
        placed in the ``plot_dir`` directory, as determined by the
        config file (see below)
    (6) ``opt_elem.html`` - a historgram showing the number of datasets
        for each filter, placed in the ``plot_dir`` directory, as
        determined by the config file (see below)
    (7) ``<dataset name>_periodgram.png`` - Lomb-Scargle periodograms
        for each dataset (both individual and composite), placed in the
        ``plot_dir`` directory, as determined by the config file (see
        below). Additionally, periodograms that are deemed interesting
        are saved in a separate ``periodogram_subset`` directory under
        the ``plot_dir`` directory.
    (8) a log file in the ``log_dir`` directory as determined by the
        config file (see below)

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``plot_dir`` - The path to where hstlc output plots are
          stored
        - ``composite_dir`` - The path to where hstlc composite output
          products are stored
        - ``log_dir`` - The path to where the log file will be stored
        - ``num_cores`` - The number of cores to use during
          multiprocessing

    Other external library dependencies include:
        - ``astropy``
        - ``bokeh``
        - ``lightcurve_pipeline``
        - ``matplotlib``
        - ``numpy``
        - ``pymysql``
        - ``sqlalchemy``
"""

from collections import Counter
from collections import OrderedDict
import glob
import itertools
import logging
import multiprocessing as mp
import os
import platform

from astropy.io import fits
from astropy.table import Table
import bokeh
from bokeh import charts
from bokeh.io import gridplot
from bokeh.plotting import figure
from bokeh.embed import components
from bokeh import palettes
import numpy as np

import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
plt.rcParams['axes.formatter.useoffset'] = False

#import seaborn as sns
#sns.set(style="dark")

from lightcurve_pipeline.utils.periodogram_stats import get_periodogram_stats
from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.utils import set_permissions
from lightcurve_pipeline.utils.utils import setup_logging
from lightcurve_pipeline.database.database_interface import engine
from lightcurve_pipeline.database.database_interface import session
from lightcurve_pipeline.database.database_interface import Metadata
from lightcurve_pipeline.database.database_interface import Stats

#-------------------------------------------------------------------------------

def bar_opt_elem():
    """
    Create a bar chart showing the number of composite lightcurves
    for each COS & STIS optical element
    """

    logging.info('Creating optical element bar chart')

    # Query for data
    query = session.query(Metadata.instrume, Metadata.opt_elem).all()
    instrumes = [result[0] for result in query]
    opt_elems = [result[1] for result in query]
    opt_elems_set = sorted(list(set(opt_elems)))

    # Initialize dictionaries that store all optical elements
    cos_dict, stis_dict = {}, {}
    for opt_elem in opt_elems_set:
        cos_dict[opt_elem] = 0
        stis_dict[opt_elem] = 0

    # Count number of opt_elems
    for instrument, opt_elem in zip(instrumes, opt_elems):
        if instrument == 'COS':
            cos_dict[opt_elem] += 1
        elif instrument == 'STIS':
            stis_dict[opt_elem] += 1

    # Determine plotting values
    cat = list(opt_elems_set)
    xyvalues = OrderedDict()
    xyvalues['COS'] = [cos_dict[opt_elem] for opt_elem in opt_elems_set]
    xyvalues['STIS'] = [stis_dict[opt_elem] for opt_elem in opt_elems_set]

    # Make plots
    bar = charts.Bar(xyvalues,
        cat,
        xlabel='Optical Element',
        ylabel='# of Lightcurves',
        stacked=True,
        legend = 'top_right')
    bar.background_fill = '#cccccc'
    bar.outline_line_color = 'black'
    charts.output_file(os.path.join(SETTINGS['plot_dir'], 'opt_elem.html'))
    plot_file = os.path.join(SETTINGS['plot_dir'], 'opt_elem.html')
    charts.save(obj=bar, filename=plot_file)
    set_permissions(plot_file)

#-------------------------------------------------------------------------------

def configuration_piechart():
    """
    Create a piechart showing distribution of configurations for each
    imstrument/detector combination
    """

    logging.info('Creating target piechart')

    data_dir = SETTINGS['composite_dir']

    configs = {}
    for dataset in glob.glob(os.path.join(data_dir, '*.fits')):

        instrument = os.path.split(dataset)[1].split('_')[3]
        grating = os.path.split(dataset)[1].split('_')[5]
        cenwave = os.path.split(dataset)[1].split('_')[6]

        if not instrument in configs:
            configs[instrument] = ['{}/{}'.format(grating, cenwave)]
        else:
            configs[instrument].append('{}/{}'.format(grating,cenwave))

    # COS FUV
    try:
        settings = Counter(configs['cos-fuv'])
        charts.output_file(os.path.join(SETTINGS['plot_dir'], 'pie_config_cos_fuv.html'))
        plot = charts.Donut(settings.values(),
                            settings.keys(),
                            width=1200,
                            height=600,
                            title='COS FUV breakdown')
        plot_file = os.path.join(SETTINGS['plot_dir'], 'pie_config_cos_fuv.html')
        charts.save(obj=plot, filename=plot_file)
        set_permissions(plot_file)
    except KeyError:
        logging.info('No COS FUV datasets')

    # COS NUV
    try:
        settings = Counter(configs['cos-nuv'])
        charts.output_file(os.path.join(SETTINGS['plot_dir'], 'pie_config_cos_nuv.html'))
        plot = charts.Donut(settings.values(),
                            settings.keys(),
                            width=1200,
                            height=600,
                            title='COS NUV breakdown')
        plot_file = os.path.join(SETTINGS['plot_dir'], 'pie_config_cos_nuv.html')
        charts.save(obj=plot, filename=plot_file)
        set_permissions(plot_file)
    except KeyError:
        logging.info('No COS NUV datasets')

    # STIS FUV
    try:
        settings = Counter(configs['stis-fuv-mama'])
        charts.output_file(os.path.join(SETTINGS['plot_dir'], 'pie_config_stis_fuv.html'))
        plot = charts.Donut(settings.values(),
                            settings.keys(),
                            width=1200,
                            height=600,
                            title='STIS FUV breakdown')
        plot_file = os.path.join(SETTINGS['plot_dir'], 'pie_config_stis_fuv.html')
        charts.save(obj=plot, filename=plot_file)
        set_permissions(plot_file)
    except KeyError:
        logging.info('No STIS FUV-MAMA datasets')

    # STIS NUV
    try:
        settings = Counter(configs['stis-nuv-mama'])
        charts.output_file(os.path.join(SETTINGS['plot_dir'], 'pie_config_stis_nuv.html'))
        plot = charts.Donut(settings.values(),
                            settings.keys(),
                            width=1200,
                            height=600,
                            title='STIS NUV breakdown')
        plot_file = os.path.join(SETTINGS['plot_dir'], 'pie_config_stis_nuv.html')
        charts.save(obj=plot, filename=plot_file)
        set_permissions(plot_file)
    except KeyError:
        logging.info('No STIS NUV datasets')


#-------------------------------------------------------------------------------

def dataset_dashboard(filename, plot_file=''):
    """
    Creates interactive bokeh 'dashboard' plot for the given filename

    Parameters
    ----------
    filename : str
        The path to the lightcurve
    plot_file : str
        The path to the PNG plot.  The user can supply this argument if
        they wish to update the plot or save to a specific location.
    """

    logging.info('Creating bokeh dashboard plots for {}'.format(filename))

    if not plot_file:
        plot_file = filename.replace('.fits', '.html')
    if os.path.exists(plot_file):
        os.remove(plot_file)

    bokeh.io.output_file(plot_file)
    TOOLS = 'pan,wheel_zoom,box_zoom,reset,resize,box_select,lasso_select,save'

    with fits.open(filename) as hdu:
        source = bokeh.models.ColumnDataSource(data={col : hdu[1].data[col] for col in hdu[1].data.names})

        endless_colors = itertools.cycle(palettes.Spectral6)
        colors = [endless_colors.next() for i in np.unique(hdu[1].data['dataset'])]

        dset_counts = Counter(hdu[1].data['dataset'])
        repeats = [dset_counts[key] for key in sorted(dset_counts.keys())]

        colors = np.repeat(colors, repeats)

        axes = []
        for key in ['gross', 'net', 'flux', 'error', 'background']:
            if len(axes) == 0:
                axes.append(figure(tools=TOOLS, plot_width=900, plot_height=350, title=key, toolbar_location='right'))
            else:
                axes.append(figure(tools=TOOLS, x_range=axes[0].x_range, plot_width=900, plot_height=350, title=key, toolbar_location='right'))
            axes[-1].circle('mjd',
                            key,
                            source=source,
                            size=12,
                            color=colors,
                            fill_alpha=1)


    # put all the plots in a grid layout
    p = bokeh.io.vplot(*axes)

    charts.save(obj=p, filename=plot_file)
    charts.reset_output()
    del p
    set_permissions(plot_file)

#-------------------------------------------------------------------------------

def exploratory_tables():
    """
    Create html tables containing data from the stats table as well as
    plots, broken down into interesting, boring, and null results
    """

    # Interesting results
    logging.info('Creating exploratory table for interesting datasets')
    interesting_query = session.query(Stats.lightcurve_path, Stats.lightcurve_filename).\
        filter(Stats.poisson_factor != 'NULL').\
        filter(Stats.poisson_factor >= 1.2).\
        filter(Stats.lightcurve_path.like('%composite%')).all()
    interesting_datasets = [os.path.join(item[0], item[1]) for item in interesting_query]
    make_exploratory_table(interesting_datasets, os.path.join(SETTINGS['plot_dir'], 'interesting_hstlc.html'))

    # Boring results
    logging.info('Creating exploratory table for boring datasets')
    boring_query = session.query(Stats.lightcurve_path, Stats.lightcurve_filename).\
        filter(Stats.poisson_factor != 'NULL').\
        filter(Stats.poisson_factor < 1.2).\
        filter(Stats.lightcurve_path.like('%composite%')).all()
    boring_datasets = [os.path.join(item[0], item[1]) for item in boring_query]
    make_exploratory_table(boring_datasets, os.path.join(SETTINGS['plot_dir'], 'boring_hstlc.html'))

    # NULL results
    logging.info('Creating exploratory table for NULL datasets')
    null_query = session.query(Stats.lightcurve_path, Stats.lightcurve_filename).\
        filter(Stats.poisson_factor == 'NULL').\
        filter(Stats.lightcurve_path.like('%composite%')).all()
    null_datasets = [os.path.join(item[0], item[1]) for item in null_query]
    make_exploratory_table(null_datasets, os.path.join(SETTINGS['plot_dir'], 'null_hstlc.html'))

#-------------------------------------------------------------------------------

def histogram_exptime():
    """
    Create a histogram showing the distribution of exposure times for
    the composite lightcurves
    """

    logging.info('Creating exposure time historgram')

    data_dir = SETTINGS['composite_dir']

    exptime_data = {}
    for dataset in glob.glob(os.path.join(data_dir, '*.fits')):
        with fits.open(dataset) as hdu:
            targname = os.path.split(dataset)[-1].split('_')[4]
            exptime = hdu[0].header['EXPTIME']

            if targname in exptime_data:
                exptime_data[targname] += exptime
            else:
                exptime_data[targname] = exptime

    charts.output_file(os.path.join(SETTINGS['plot_dir'], 'exptime_histogram.html'))

    times = np.array(exptime_data.values())
    names = np.array(exptime_data.keys())
    indx = np.argsort(times)

    times = list(times[indx])[::-1][:30]
    names = list(names[indx])[::-1][:30]

    bar = charts.Bar(times,
                     cat=names,
                     xlabel='Target',
                     ylabel='Exptime (s)',
                     width=700,
                     height=400,
                     title='Cumulative Exptime per Target')

    bar.background_fill = '#cccccc'
    bar.outline_line_color = 'black'

    # change just some things about the x-grid
    #bar.xgrid.grid_line_color = None

    # change just some things about the y-grid
    #bar.ygrid.band_fill_alpha = 0.1
    #bar.ygrid.band_fill_color = "navy"

    #bar.toolbar_location = None

    #charts.show(bar)

    script, div = components(bar)

    plot_file = os.path.join(SETTINGS['plot_dir'], 'exptime_histogram.html')
    charts.save(obj=bar, filename=plot_file)
    set_permissions(plot_file)

#-------------------------------------------------------------------------------

def make_exploratory_table(dataset_list, table_name):
    """
    Create html tables containing data from the stats table as well as
    plots

    Parameters
    ----------
    dataset_list : list
        A list of the paths to the datasets to process
    table_name : str
        The path to the output file
    """

    info = []

    if len(dataset_list) > 0:

        for dataset in dataset_list:
            path, name = os.path.split(dataset)

            with fits.open(dataset) as hdu:
                exptime = hdu[0].header['EXPTIME']

            targname = os.path.split(dataset)[1].split('_')[4]
            instrument = os.path.split(dataset)[1].split('_')[3].split('-')[0]
            grating = os.path.split(dataset)[1].split('_')[5]
            cenwave = os.path.split(dataset)[1].split('_')[6]
            aperture = os.path.split(dataset)[1].split('_')[7]

            results = session.query(Stats).filter(Stats.lightcurve_filename == name).all()
            total = results[0].total
            mean = results[0].mean
            poisson_f = results[0].poisson_factor
            pearson_r = results[0].pearson_r
            pearson_p = results[0].pearson_p

            plot_name = dataset.replace('.fits', '.html')
            plot_name_static = dataset.replace('.fits', '.png')

            plot_html = """<a href="{}" target="_blank"><img width="400" src="{}"><a>""".format(plot_name, plot_name_static)
            new_row = (targname, plot_html, instrument, grating, cenwave, aperture, exptime, total, mean, poisson_f, pearson_r, pearson_p, dataset)
            info.append(new_row)

        out_tab = Table(rows=info,
                        names=('target', 'plot', 'instrument', 'grating', 'cenwave', 'aperture', 'exptime', 'total', 'mean', 'poisson_f', 'pearson_r', 'pearson_p', 'filename'))
        out_tab.write(table_name, format='jsviewer')

        if platform.system() == 'Linux':
            os.system("""sed -i "s/&lt;/</g" {}""".format(table_name))
            os.system("""sed -i "s/&gt;/>/g" {}""".format(table_name))
        elif platform.system() == 'Darwin':
            os.system("""sed -i '' "s/&lt;/</g" {}""".format(table_name))
            os.system("""sed -i '' "s/&gt;/>/g" {}""".format(table_name))
        else:
            raise ValueError("OS {} is not supported for this operation".format(platform.system()))

#-------------------------------------------------------------------------------

def periodogram(dataset):
    """
    Create a Lomb-Scargle periodgram for the given dataset

    Parameters
    ----------
    dataset : string
        The path to the dataset
    """

    # Get the data
    hdulist = fits.open(dataset, mode='readonly')
    counts = hdulist[1].data['net']
    times = hdulist[1].data['mjd']

    # Make sure the byte-order is native
    counts = counts.byteswap().newbyteorder()
    times = times.byteswap().newbyteorder()

    # Define frequency space (in days)
    short_freq = (hdulist[0].header['STEPSIZE'] / (60. * 60. * 24.)) # Step size
    med_freq = (10. / (60. * 24.)) # 10 minutes
    long_freq = 1. / 24. # 1 hour
    max_freq = 10. / 24. # 10 hours

    # Get the periodogram information
    short_periods, short_power, short_mean, short_three_sigma, short_significant_periods, short_significant_powers = get_periodogram_stats(dataset, 'short')
    med_periods, med_power, med_mean, med_three_sigma, med_significant_periods, med_significant_powers = get_periodogram_stats(dataset, 'med')
    long_periods, long_power, long_mean, long_three_sigma, long_significant_periods, long_significant_powers = get_periodogram_stats(dataset, 'long')

    # Make plot if there are significant peaks above the threshold
    significant_threshold = 0.30
    all_significant_powers = list(short_significant_powers) + list(med_significant_powers) + list(long_significant_powers)
    if len(all_significant_powers) > 0:
        max_significant_power = max(all_significant_powers)
        if max_significant_power >= significant_threshold:

            logging.info('Creating periodogram for dataset {}'.format(dataset))

            fig = plt.figure(figsize=(10, 8))
            ax1 = fig.add_subplot(411)
            ax2 = fig.add_subplot(412)
            ax3 = fig.add_subplot(413)
            ax4 = fig.add_subplot(414)
            ax1.minorticks_on()

            # Top panel - the data
            ax1.plot(times, counts, 'b+')
            ax1.set(xlabel='MJD', ylabel='Net')

            # 2nd panel - short frequency space
            ax2.plot(short_periods, short_power)
            ax2.axhline(short_mean, color='r', linestyle='-')
            ax2.axhline(short_three_sigma, color='g', linestyle='-')
            for period in short_significant_periods:
                ax2.axvline(period, color='k', linestyle='--')
            ax2.set(xlim=(short_freq, med_freq), title='Short Frequency Space')

            # 3rd panal - medium frequency space
            ax3.plot(med_periods, med_power)
            ax3.axhline(med_mean, color='r', linestyle='-')
            ax3.axhline(med_three_sigma, color='g', linestyle='-')
            for period in med_significant_periods:
                ax3.axvline(period, color='k', linestyle='--')
            ax3.set(xlim=(med_freq, long_freq), ylabel='Lomb-Scargle Power', title='Medium Frequency Space')

            # bottom panel - long frequency space
            ax4.plot(long_periods, long_power)
            ax4.axhline(long_mean, color='r', linestyle='-')
            ax4.axhline(long_three_sigma, color='g', linestyle='-')
            for period in long_significant_periods:
                ax4.axvline(period, color='k', linestyle='--')
            ax4.set(xlim=(long_freq, max_freq), xlabel='Period (Days)', title='Long Frequency Space')

            # Save the plot
            fig.tight_layout()
            filename = '{}_periodogram.png'.format(os.path.basename(dataset).split('_curve.fits')[0])
            save_loc = os.path.join(SETTINGS['plot_dir'], 'periodogram_subset', filename)
            plt.savefig(save_loc)
            plt.close()
            set_permissions(save_loc)

#-------------------------------------------------------------------------------

def plot_dataset(filename, plot_file=''):
    """
    Create an interactive bokeh lightcurve plot for the given filename

    Parameters
    ----------
    filename : str
        The path to the lightcurve
    plot_file : str
        The path to the PNG plot.  The user can supply this argument if
        they wish to update the plot or save to a specific location
    """

    logging.info('Creating bokeh lightcurve for {}'.format(filename))

    path, name = os.path.split(filename)

    if not plot_file:
        plot_file = name.replace('.fits', '.html')
    if os.path.exists(plot_file):
        os.remove(plot_file)

    TOOLS = 'pan,wheel_zoom,box_zoom,box_select,lasso_select,reset,resize,save'

    charts.output_file(plot_file)
    p = figure(tools=TOOLS, toolbar_location='above', logo='grey', plot_width=700)
    p.background_fill= "#cccccc"

    with fits.open(filename) as hdu:
        p.circle(hdu[1].data['MJD'],
                 hdu[1].data['FLUX'],
                 size=12,
                 line_color="black",
                 fill_alpha=0.8)

    p.xaxis.axis_label='Time (MJD)'
    p.yaxis.axis_label='Net (cnts/sec)'
    p.grid.grid_line_color='white'

    charts.save(obj=p, filename=plot_file)
    charts.reset_output()
    del p
    set_permissions(plot_file)

#-------------------------------------------------------------------------------

def plot_dataset_static(filename, plot_file=''):
    """
    Creates static PNG lightcurve plot for the given filename

    Parameters
    ----------
    filename : str
        The path to the lightcurve
    plot_file : str
        The path to the PNG plot.  The user can supply this argument if
        they wish to update the plot or save to a specific location
    """

    logging.info('Creating static PNG for {}'.format(filename))

    if not plot_file:
        plot_file = filename.replace('.fits', '.png')
    if os.path.exists(plot_file):
        os.remove(plot_file)

    fig = plt.figure(figsize=(10, 1))
    ax = fig.add_subplot(1, 1, 1)

    with fits.open(filename) as hdu:
        indx = np.argsort(hdu[1].data['MJD'])
        xvals = np.arange(len(hdu[1].data['MJD']))
        yvals = hdu[1].data['FLUX']

        try:
            colors = hdu[1].data['dataset'][indx]
        except KeyError:
            colors = np.ones(xvals.shape)

        ax.scatter(xvals,
                   yvals[indx],
                   c=colors,
                   marker='o')

    ax.set_xlim(0, xvals.max())
    ax.set_ylim(yvals[indx].min(), yvals[indx].max())
    fig.savefig(plot_file, bbox_inches='tight')
    plt.close(fig)
    del fig
    set_permissions(plot_file)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

def main():
    """The main function of the ``make_hstlc_plots`` script
    """

    # Configure logging
    module = os.path.basename(__file__).strip('.py')
    setup_logging(module)

    # Make matplotlib and bokeh lightcurve plots
    composite_datasets = glob.glob(os.path.join(SETTINGS['composite_dir'], '*.fits'))
    logging.info('Creating matplotlib and bokeh lightcurve plots for {} datasets using {} cores'.format(len(composite_datasets), SETTINGS['num_cores']))
    pool = mp.Pool(processes=SETTINGS['num_cores'])
    pool.map(plot_dataset_static, composite_datasets)
    pool.map(dataset_dashboard, composite_datasets)
    pool.close()
    pool.join()

    # Make exploratory tables
    exploratory_tables()

    # Make exposure time histogram
    histogram_exptime()

    # Make configuration piechart
    configuration_piechart()

    # Make optical element bar chart
    bar_opt_elem()

    # Make individual bokeh plots for some examples
    #plot_dataset('/grp/hst/hstlc/hst13902/outputs/composite/V4046SGR_FUV_G130M_1300_curve.fits', os.path.join(SETTINGS['plot_dir'], 'example_flare.html'))
    #plot_dataset('/grp/hst/hstlc/hst13902/outputs/composite/IR-COM_FUV_G140L_1105_curve.fits', os.path.join(SETTINGS['plot_dir'], 'example_transit.html'))

    # Make periodograms
    results = session.query(Stats.lightcurve_path, Stats.lightcurve_filename).\
        filter(Stats.periodogram == True).all()
    datasets = [os.path.join(item.lightcurve_path, item.lightcurve_filename) for item in results]
    logging.info('Making periodograms for {} datasets over {} cores'.format(len(datasets), SETTINGS['num_cores']))
    pool = mp.Pool(processes=SETTINGS['num_cores'])
    pool.map(periodogram, datasets)
    pool.close()
    pool.join()

    logging.info('Processing complete')

#-------------------------------------------------------------------------------

if __name__ == '__main__':

    main()



"""
This module contains functions that create lightcurves (in the form of
FITS tables) from both individual and composite datasets.  It is
essentially a wrapper around the ``lightcurve`` library, which makes
lightcurve objects, for example:

::

    lc = lightcurve.LightCurve(filename)

This module uses multiprocessing to process the composite lightcurves
over numerous cores, as given by the ``num_cores`` key in the config
file (see below).

**Authors:**

    Matthew Bourque

**Use:**

    This module is intended to be imported from the ``ingest_hstlc``
    script as such:

::

    from lightcurve_pipeline.ingest.make_lightcurves import make_composite_lightcurves
    from lightcurve_pipeline.ingest.make_lightcurves import make_individual_lightcurve

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``ingest_dir`` - The path to where files to be ingested are
          stored
        - ``composite_dir`` - The path to where hstlc composite output
          products are stored
        - ``num_cores`` - The number of cores to use during
          multiprocessing

    Other external library dependencies include:
        - ``pymysql``
        - ``sqlalchemy``
        - ``lightcurve``
        - ``lightcurve_pipeline``
"""

import logging
import multiprocessing
import os
import traceback

import lightcurve

from lightcurve_pipeline.database.database_interface import get_session
from lightcurve_pipeline.database.database_interface import Metadata
from lightcurve_pipeline.database.database_interface import Outputs
from lightcurve_pipeline.utils.utils import make_directory
from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.utils import set_permissions

# -----------------------------------------------------------------------------

def make_composite_lightcurves():
    """Create composite lightcurves made up of datasets with similar
    ``targname``, ``detector``, ``opt_elem``, ``cenwave``, and
    ``aperture``.  The lightcurves that require processing are
    determined by ``NULL`` ``composite_path`` values in the ``outputs``
    table of the database.
    """

    logging.info('')
    logging.info('Creating composite lightcurves')

    # Create composite directory if it doesn't already exist
    make_directory(SETTINGS['composite_dir'])

    # Get list of datasets that need to be (re)processed by querying
    # for empty composite records
    session = get_session()
    datasets = session.query(Metadata.instrume, Metadata.detector,
        Metadata.targname, Metadata.opt_elem, Metadata.cenwave,
        Metadata.aperture).join(Outputs)\
        .filter(Outputs.composite_path == None).all()
    datasets = set(datasets)
    session.close()

    # Process each dataset using multiprocessing
    logging.info('Creating {} composites using {} core(s)'.format(
        len(datasets), SETTINGS['num_cores']))
    logging.info('')
    pool = multiprocessing.Pool(processes=SETTINGS['num_cores'])
    pool.map(process_dataset, datasets)
    pool.close()
    pool.join()

# -----------------------------------------------------------------------------

def make_individual_lightcurve(metadata_dict, outputs_dict):
    """Create a lightcurve for an individual dataset

    Parameters
    ----------
    metadata_dict : dict
        A dictionary containing metadata of the file
    outputs_dict : dict
        A dictionary containing output product information

    Returns
    -------
    success : bool
        True/False status of successful extraction
    """

    success = True

    # Create parent output directory if necessary
    make_directory(outputs_dict['individual_path'])

    # Create the lightcurve if it doesn't already exist
    outputname = os.path.join(outputs_dict['individual_path'],
        outputs_dict['individual_filename'])

    if not os.path.exists(outputname):
        inputname = os.path.join(SETTINGS['ingest_dir'],
            metadata_dict['filename'])

        try:
            lc = lightcurve.LightCurve(filename=inputname, step=2, verbosity=1)
            lc.write(outputname)
            set_permissions(outputname)
        except Exception as e:
            logging.warn('Exception raised for {}'.format(outputname))
            logging.warn('\t{}'.format(e.message))
            success = False

    return success

# -----------------------------------------------------------------------------

def process_dataset(dataset):
    """Create a composite lightcurve for the given dataset

    Parameters
    ----------
    dataset : list
        A list comprised of four elements.  The first element is the
        ``targname``, the second is the ``detector``, the third is the
        ``opt_elem``, and the fourth is the ``cenwave``.
    """

    try:

        # Parse the dataset information
        instrume = dataset[0]
        detector = dataset[1]
        targname = dataset[2]
        opt_elem = dataset[3]
        cenwave = dataset[4]
        aperture = dataset[5]

        # Get list of files for each dataset to be processed
        session = get_session()
        filelist = session.query(
            Metadata.id, Metadata.path, Metadata.filename)\
            .filter(Metadata.instrume == instrume)\
            .filter(Metadata.detector == detector)\
            .filter(Metadata.targname == targname)\
            .filter(Metadata.opt_elem == opt_elem)\
            .filter(Metadata.cenwave == cenwave)\
            .filter(Metadata.aperture == aperture).all()
        metadata_ids = [item[0] for item in filelist]
        files_to_process = [os.path.join(item[1], item[2]) for item in filelist]
        logging.info('Processing dataset: {}\t{}\t{}\t{}\t{}\t{}: {} files to process'.format(
            instrume, detector, targname, opt_elem, cenwave, aperture,
            len(files_to_process)))
        session.close()

        # Perform the extraction
        path = SETTINGS['composite_dir']
        output_filename = 'hlsp_hstlc_hst_{}-{}_{}_{}_{}_{}_v1_sci.fits'.format(
            instrume, detector, targname, opt_elem, cenwave, aperture).lower()
        save_loc = os.path.join(path, output_filename)
        lightcurve.composite(files_to_process, save_loc, step=2)
        set_permissions(save_loc)
        logging.info('\tComposite lightcurve saved to {}'.format(save_loc))

        # Update the outputs table with the composite information
        session = get_session()
        for metadata_id in metadata_ids:
            session.query(Outputs)\
                .filter(Outputs.metadata_id == metadata_id)\
                .update({'composite_path':path,
                    'composite_filename':output_filename})
        session.commit()
        session.close()

    # Track any errors that happen during processing
    except Exception as error:
        dataset_name = 'hlsp_hstlc_hst_{}-{}_{}_{}_{}_{}_curve.fits'.format(
            dataset[0], dataset[1], dataset[2], dataset[3], dataset[4],
            dataset[5])
        trace = 'Failed to create composite for dataset {}\n{}'.format(
            dataset_name, traceback.format_exc())
        logging.critical(trace)



#! /usr/bin/env python

"""
Generate the ``stats`` table in the hstlc database, which stores
statistics for every lightcurve (both individual and composite).
Several statistics and flags are computed for every lightcurve
generated by ingest_hstlc, including statistics on counts, primitive
data quality statistics, and flags to indicate interesting datasets.

**Authors:**

    Matthew Bourque

**Use:**

    This script is intended to be executed as part of the
    hstlc_pipeline shell script.  However, users can also execute this
    script via the command line as such:

    >>> build_stats_table [product_type]

    ``product_type`` (*required*) - The type of lightcurves to process.
    Can be *individual* for individual lightcurves, *composite* for
    composite lightcurves, or *both* for both types.

**Outputs:**

    (1) New and/or updated entries in the ``stats`` table in the hstlc
        database
    (2) a log file in the ``log_dir`` directory as determined by the
        config file (see below)

**Dependencies:**

    (1) Users must have access to the hstlc database.

    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string
        - ``log_dir`` - The path to where the log file will be stored

    Other external library dependencies include:
        - ``astropy``
        - ``lightcurve``
        - ``lightcurve_pipeline``
        - ``numpy``
        - ``pymysql``
        - ``scipy``
        - ``sqlalchemy``
"""

import argparse
import logging
import os

from astropy.io import fits
import lightcurve
import numpy as np
from scipy import signal
from scipy.stats.stats import pearsonr

from lightcurve_pipeline.utils.periodogram_stats import get_periodogram_stats
from lightcurve_pipeline.utils.utils import insert_or_update
from lightcurve_pipeline.utils.utils import SETTINGS
from lightcurve_pipeline.utils.utils import setup_logging
from lightcurve_pipeline.database.database_interface import engine
from lightcurve_pipeline.database.database_interface import session
from lightcurve_pipeline.database.database_interface import Stats
from lightcurve_pipeline.database.database_interface import Outputs
from lightcurve_pipeline.database.update_database import update_stats_table

# -----------------------------------------------------------------------------

def get_lightcurves(product_type):
    """Queries the ``outputs`` table to build a list of lightcurves to
    get stats from

    Parameters
    ----------
    product_type : string
        The type of lightcurves to process. Can either be
        ``individual``, ``composite``, or ``both``

    Returns
    -------
    lightcurves : list
        A list of paths to lightcurve products
    """

    # Build lists of paths to lightcurves
    results = session.query(Outputs).all()
    individual_lightcurves = set([os.path.join(result.individual_path, result.individual_filename) for result in results])
    composite_lightcurves = set([os.path.join(result.composite_path, result.composite_filename) for result in results if result.composite_filename != None])

    # Return appropriate list based on product_type
    if product_type == 'individual':
        lightcurves = individual_lightcurves
    elif product_type == 'composite':
        lightcurves = composite_lightcurves
    elif product_type == 'both':
        lightcurves = individual_lightcurves | composite_lightcurves

    return lightcurves

# -----------------------------------------------------------------------------

def get_stats(dataset):
    """Gathers various statistics for the given lightcurve product

    Parameters
    ----------
    dataset : string
        The path to the lightcurve product

    Returns
    -------
    stats_dict : dictionary
        A dictionary whose keys are column names of the ``stats``
        table and whose values are the corresponding statistics
    """

    # Initialize dictionary
    stats_dict = {}

    # Add path information
    stats_dict['lightcurve_path'] = os.path.dirname(dataset)
    stats_dict['lightcurve_filename'] = os.path.basename(dataset)

    # Open lightcurve and extract information
    hdulist = fits.open(dataset, mode='readonly')
    counts = hdulist[1].data.counts
    mjd = hdulist[1].data.mjd

    # Populate stats_dict with counts information
    stats_dict['total'] = int(np.sum(counts))

    # If total counts is zero, there is nothing else to compute
    # If there are counts, then populate the rest of stats_dict
    if stats_dict['total'] > 0:

        # Populate stats dict with count stats
        stats_dict['mean'] = float(np.mean(counts))
        stats_dict['mu'] = float(np.sqrt(stats_dict['mean']))
        stats_dict['stdev'] = float(np.std(counts))

        # Populate stats_dict with poisson information
        try:
            stats_dict['poisson_factor'] = float(stats_dict['stdev'] / stats_dict['mu'])
        except ZeroDivisionError:
            stats_dict['poisson_factor'] = 9999999.

        # Populate stats_dict with pearson information
        pearson_results = pearsonr(mjd, counts)
        stats_dict['pearson_r'] = float(pearson_results[0])
        stats_dict['pearson_p'] = float(pearson_results[1])

        # Set 'interesting periodogram' flag
        interesting_periodogram = False
        significant_threshold = 0.30
        for freq_space in ['short', 'med', 'long']:
            periods, power, mean, three_sigma, significant_periods, significant_powers = get_periodogram_stats(dataset, freq_space)
            if len(significant_powers) > 0:
                if max(significant_powers) >= significant_threshold:
                    interesting_periodogram = True
        stats_dict['periodogram'] = interesting_periodogram

    return stats_dict

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

def parse_args():
    """Parse command line arguments

    Returns
    -------
    args : argparse object
        An argparse object containing the arguments
    """

    product_help = ('The type of output products to process.  Can be '
        '"individual", "composite", or "both".')

    parser = argparse.ArgumentParser()
    parser.add_argument('product_type', action='store', type=str, help=product_help)
    args = parser.parse_args()

    # Make sure the argument is a valid option
    valid_options = ['individual', 'composite', 'both']
    explanation = '{} is not a valid option.  Please choose "individual", "composite", or "both".'.format(args.product_type)
    assert args.product_type in valid_options, explanation

    return args

# -----------------------------------------------------------------------------

def main():
    """The main function of the ``build_stats_table`` script
    """

    # Configure logging
    module = os.path.basename(__file__).strip('.py')
    setup_logging(module)

    # Parse arguments
    args = parse_args()

    # Query the outputs table for a list of lightcurves
    lightcurves = get_lightcurves(args.product_type)

    # For each lightcurve, compute statistics and update the database
    logging.info('{} datasets to process'.format(len(lightcurves)))
    for dataset in lightcurves:
        logging.info('Processing {}'.format(dataset))
        stats_dict = get_stats(dataset)
        update_stats_table(stats_dict, dataset)

    logging.info('Processing complete')

# -----------------------------------------------------------------------------

if __name__ == '__main__':

    main()



"""
This module serves as the interface and connection module to the hstlc
database.  The ``load_connection()`` function within allows the user
to conenct to the database via the ``session``, ``base``, and
``engine`` objects (described below).  The classes within serve as the
object-relational mappings (ORMs) that define the individual tables of
the database, and are used to build the tables via the ``base`` object.

The ``engine`` object serves as the low-level database API and perhaps
most importantly contains dialects which allows the sqlalchemy module
to communicate with the database.

The ``base`` object serves as a base class for class definitions.  It
produces ``Table`` objects and constructs ORMs.

The ``session`` object manages operations on ORM-mapped objects, as
construced by the ``base``.  These operations include querying, for
example.

**Authors:**

    Matthew Bourque

**Use:**

    This module is intended to be imported from various hstlc modules
    and scripts.  The objects that are importable from this module are
    as follows:

::

    from lightcurve_pipeline.database.database_interface import engine
    from lightcurve_pipeline.database.database_interface import base
    from lightcurve_pipeline.database.database_interface import session
    from lightcurve_pipeline.database.database_interface import Metadata
    from lightcurve_pipeline.database.database_interface import Outputs
    from lightcurve_pipeline.database.database_interface import BadData
    from lightcurve_pipeline.database.database_interface import Stats

**Dependencies:**

    (1) Users must have access to the hstlc database
    (2) Users must also have a ``config.yaml`` file located in the
        ``lightcurve_pipeline/utils/`` directory with the following
        keys:

        - ``db_connection_string`` - The hstlc database connection
          string

    Other external library dependencies include:
        - ``pymysql``
        - ``sqlalchemy``
        - ``lightcurve_pipeline``
"""

import pymysql
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Enum
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

from lightcurve_pipeline.utils.utils import SETTINGS

# -----------------------------------------------------------------------------

def get_session():
    """Return the ``session`` object of the database connection

    In many cases, all that is needed is the ``session`` object to
    interact with the database.  This function can be used just to
    establish a connection and retreive the ``session`` object.

    Returns
    -------
    session : sqlalchemy.orm.session.Session
        Provides a holding zone for all objects loaded or associated
        with the database.
    """

    session, base, engine = load_connection(SETTINGS['db_connection_string'])

    return session

# -----------------------------------------------------------------------------

def load_connection(connection_string, echo=False):
    """Create and return a connection to the database given in the
    connection string.

    Parameters
    ----------
    connection_string : str
        A string that points to the database conenction.  The
        connection string is in the following form:
        ``dialect+driver://username:password@host:port/database``
    echo : bool
        Show all SQL produced

    Returns
    -------
    session : sesson object
        Provides a holding zone for all objects loaded or associated
        with the database.
    base : base object
        Provides a base class for declarative class definitions.
    engine : engine object
        Provides a source of database connectivity and behavior.
    """

    engine = create_engine(connection_string, echo=echo)
    base = declarative_base(engine)
    Session = sessionmaker(bind=engine)
    session = Session()

    return session, base, engine

session, base, engine = load_connection(SETTINGS['db_connection_string'])

# -----------------------------------------------------------------------------
# Define ORMs
# -----------------------------------------------------------------------------

class Metadata(base):
    """ORM for the metadata table"""

    __tablename__ = 'metadata'
    id = Column(Integer(), nullable=False, primary_key=True)
    filename = Column(String(30), unique=True, nullable=False, index=True)
    path = Column(String(100), nullable=False)
    ingest_date = Column(Date, nullable=False)
    telescop = Column(String(10), nullable=False)
    instrume = Column(String(10), nullable=False)
    targname = Column(String(30), nullable=False)
    cal_ver = Column(String(30), nullable=False)
    obstype = Column(String(30), nullable=False)
    cenwave = Column(Integer(), nullable=False)
    aperture = Column(String(30), nullable=False)
    detector = Column(String(30), nullable=False)
    opt_elem = Column(String(30), nullable=False)
    fppos = Column(Integer(), nullable=False)


class Outputs(base):
    """ORM for the outputs table"""
    __tablename__ = 'outputs'
    id = Column(Integer(), nullable=False, primary_key=True)
    metadata_id = Column(Integer(), ForeignKey('metadata.id'), nullable=False,
        unique=True)
    individual_path = Column(String(100))
    individual_filename = Column(String(30))
    composite_path = Column(String(100))
    composite_filename = Column(String(100))


class BadData(base):
    """ORM for bad_data table"""
    __tablename__ = 'bad_data'
    id = Column(Integer(), nullable=False, primary_key=True)
    filename = Column(String(30), unique=True, nullable=False, index=True)
    ingest_date = Column(Date, nullable=False)
    reason = Column(Enum('Bad EXPFLAG', 'Non-linear time', 'No events', 'Singular event', 'Bad Proposal', 'Short Exposure'), nullable=False)


class Stats(base):
    """ORM for stats table"""
    __tablename__ = 'stats'
    id = Column(Integer(), nullable=False, primary_key=True)
    lightcurve_path = Column(String(100), nullable=False)
    lightcurve_filename = Column(String(100), nullable=False)
    total = Column(Integer(), nullable=False)
    mean = Column(Float(10), nullable=True)
    mu = Column(Float(10), nullable=True)
    stdev = Column(Float(10), nullable=True)
    poisson_factor = Column(Float(10), nullable=True)
    pearson_r = Column(Float(10), nullable=True)
    pearson_p = Column(Float(10), nullable=True)
    periodogram = Column(Boolean(), nullable=False, default=False)
    deliver = Column(Boolean(), nullable=False, default=False)
